from trava.ext.cv.base import CV


class SklearnCV(CV):
    """
    Dummy wrapper for sklearn cross validation.
    """

    def __init__(self, sklearn_cv):
        self._sklearn_cv = sklearn_cv

    def split(self, X, y=None, groups=None):
        return self._sklearn_cv.split(X=X, y=y, groups=groups)



from abc import ABC, abstractmethod
from typing import Optional, Callable

from trava.trava_model import TravaModel
from trava.model_info import ModelInfo


class Scorer(ABC):
    """
    Wraps the score function and makes scorer in sklearn style.

    Init parameters
    ----------
    score_func: callable
        Function to calculate a metric.
        It must accept true labels as well as predicted labels.
    needs_proba: bool
        Whether a model must calculate probabilities instead of hard labels.
    requires_raw_model: bool, requires_X_y: bool
        These two parameters are required only if you already have evaluated model
        and then want to calculate some new metrics using it.
        By the moment you want to do it both raw model and X_y data may be unloaded,
        yet TravaModel caches y_pred and y_pred_proba results. If more data is required
        for the scorer or it preforms some complex logic, you may restrict using post-evaluation
        using these two parameters.
    name: str
        Custom name for your metric, func.__name__ is taken otherwise.
    metrics_kwargs: dict
        Any additional parameters to pass to score_func
    """

    def __init__(
        self,
        score_func: Callable,
        needs_proba=False,
        requires_raw_model=False,
        requires_X_y=False,
        name: Optional[str] = None,
        **metrics_kwargs
    ):
        self._func_name = name or score_func.__name__
        self._is_other_scorer = issubclass(type(self), OtherScorer)
        self._needs_proba = needs_proba
        self._requires_raw_model = requires_raw_model
        self._requires_X_y = requires_X_y

        self._scorer = self._make_scorer(score_func=score_func, **metrics_kwargs)

    @property
    def is_other_scorer(self):
        return self._is_other_scorer

    @abstractmethod
    def _make_scorer(self, score_func: Callable, **metrics_kwargs) -> Callable:
        """
        Creates a function that calls score_func with provided true labels and predictions.
        """
        pass

    @property
    def func_name(self) -> str:
        return self._func_name

    def __call__(self, trava_model: TravaModel, for_train: bool, X, X_raw, y, **kwargs):
        if self._requires_raw_model and not trava_model.raw_model():
            raise Exception("Cannot perform eval on model {} " "because it was unloaded.".format(trava_model.model_id))

        if self._requires_X_y and (X is None or X_raw is None or y is None):
            raise Exception(
                "Cannot perform eval on model {} "
                "because data is required and was unloaded.".format(trava_model.model_id)
            )

        return self._scorer(
            model=trava_model.get_model(for_train=for_train),
            model_info=trava_model,
            for_train=for_train,
            X=X,
            X_raw=X_raw,
            y=y,
            **kwargs
        )


class OtherScorer(Scorer):
    """
    You may calculate any custom value you want based on model/input data etc.
    Such metrics are separated from a model's performance metrics.
    """

    def _make_scorer(self, score_func: Callable, **metrics_kwargs) -> Callable:
        def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            return score_func(model=model, model_info=model_info, for_train=for_train, X=X, X_raw=X_raw, y=y)

        return scorer



from abc import ABC, abstractmethod


class CV(ABC):
    """
    You may want to write custom CV subclass for your specific needs.
    Or just use SklearnCV right away with the scikit-learn CV of your choice.
    """

    @abstractmethod
    def split(self, X, y=None, groups=None):
        """
        Scikit-learn style split method. Should be self-explanatory.
        """
        pass



import os
import matplotlib.pyplot as plt
from tempfile import mkdtemp
from shutil import rmtree
from functools import wraps
from typing import List, Optional, Dict

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.model_serializer import ModelSerializer
from trava.tracker import Tracker
from trava.results_handler import ResultsHandler
from trava.scorer import Scorer


def track_if_enabled(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        tracker = args[0]
        if tracker.is_enabled:
            return func(*args, **kwargs)
        return None

    return wrapper


class TravaTracker(Tracker, ResultsHandler):
    """
    Must be subclassed to have effect. Used for experiment tracking.
    Helps you to track important stuff about a model using the desired tracking framework.
    Made as a subclass of ResultsHandler for convenience integration with Trava internals.

    Init parameters
    ----------
    scorers: List[Scorer]
        Provided scorers will be used to get metrics for tracking.
    """

    def __init__(self, scorers: List[Scorer]):

        super().__init__(scorers=scorers)

        self._enabled = True
        self._started_models: Dict[str, bool] = {}

    def add_scorers(self, scorers: List[Scorer]):
        self._scorers += scorers

    @property
    def is_enabled(self) -> bool:
        return self._enabled

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    @track_if_enabled
    def start_tracking(self, model_id: str, track_name: Optional[str] = None, nested: bool = False):
        self._started_models[model_id] = True
        self._start_tracking(model_id=model_id, track_name=track_name, nested=nested)

    @track_if_enabled
    def end_tracking(self, model_id: str):
        self._end_tracking(model_id=model_id)
        del self._started_models[model_id]

    @track_if_enabled
    def track_set_tracking_group(self, group: str):
        """
        Use to group all the following tracking as a whole.
        """
        self._track_set_tracking_group(group=group)

    @track_if_enabled
    def track_model_description(self, model_id: str, description: str):
        self._track_model_description(model_id=model_id, description=description)

    @track_if_enabled
    def track_model_init_params(self, model_id: str, params: dict):
        self._track_model_init_params(model_id=model_id, params=params)

    @track_if_enabled
    def track_fit_params(self, model_id: str, params: dict):
        self._track_fit_params(model_id=model_id, params=params)

    @track_if_enabled
    def track_predict_params(self, model_id: str, params: dict):
        self._track_predict_params(model_id=model_id, params=params)

    @track_if_enabled
    def track_metric(self, model_id: str, metric: Metric, step=None):
        self.track_metric_value(model_id=model_id, name=metric.name, value=metric.value, step=step)

    @track_if_enabled
    def track_metric_value(self, model_id: str, name: str, value, step=None):
        self._track_metric_value(model_id=model_id, name=name, value=value, step=step)

    @track_if_enabled
    def track_model_info(self, model_id: str, model):
        self._track_model_info(model_id=model_id, model=model)

    @track_if_enabled
    def track_tag(self, model_id: str, tag_key: str, tag_value):
        self._track_tag(model_id=model_id, tag_key=tag_key, tag_value=tag_value)

    @track_if_enabled
    def track_artifact(self, model_id: str, filepath):
        self._track_artifact(model_id=model_id, file_path=filepath)

    @track_if_enabled
    def track_model_artifact(self, model_id: str, model, serializer: ModelSerializer):
        tmpdir = mkdtemp()
        filepath = os.path.join(tmpdir, model_id + "_model")
        serializer.save(model=model, path=filepath)
        self.track_artifact(model_id=model_id, filepath=filepath)
        rmtree(tmpdir)

    @track_if_enabled
    def track(self, model_id: str, *args, **kwargs):
        """
        Just an abstract method to track anything you want
        """
        self._track(model_id=model_id, *args, **kwargs)  # type: ignore

    @track_if_enabled
    def track_plot(self, model_id: str, fig, filename: str):
        should_restart = self._started_models.get(model_id) is None
        # TODO: probably should be checked everywhere
        if should_restart:
            self.start_tracking(model_id=model_id)

        tmpdir = mkdtemp()
        try:
            filepath = os.path.join(tmpdir, filename + ".png")
            fig.savefig(filepath)
            self.track_artifact(model_id=model_id, filepath=filepath)
        finally:
            plt.close(fig)
            rmtree(tmpdir)

            if should_restart:
                self.end_tracking(model_id=model_id)

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        # here we can ignore provided tracker param since we expect that there is only one tracker
        # per trava, so self is equal to tracker. Looks bad yet not critical to refactor it right now.
        assert tracker == self, "Something's wrong, there can be only one tracker instance per Trava"

        for model_results in results:
            self.track_model_results(model_results=model_results)

    def track_model_results(self, model_results: ModelResult):
        self._track_metrics(model_id=model_results.model_id, metrics=model_results.test_metrics(provider=self))
        self._track_metrics(model_id=model_results.model_id, metrics=model_results.other_metrics(provider=self))

    def _track_metrics(self, model_id: str, metrics: List[Metric]):
        for metric in metrics:
            if metric.is_scalar:
                self.track_metric(model_id=model_id, metric=metric)

    # TO OVERRIDE

    def _start_tracking(self, model_id: str, track_name: Optional[str] = None, nested: bool = False):
        pass

    def _end_tracking(self, model_id: str):
        pass

    def _track_set_tracking_group(self, group: str):
        pass

    def _track_model_description(self, model_id: str, description: str):
        pass

    def _track_model_init_params(self, model_id: str, params: dict):
        pass

    def _track_fit_params(self, model_id: str, params: dict):
        pass

    def _track_predict_params(self, model_id: str, params: dict):
        pass

    def _track_metric_value(self, model_id: str, name: str, value, step=None):
        pass

    def _track_model_info(self, model_id: str, model):
        pass

    def _track_tag(self, model_id: str, tag_key: str, tag_value):
        pass

    def _track_artifact(self, model_id: str, file_path):
        pass

    def _track(self, model_id: str, *args, **kwargs):
        pass

    ###



class RawDataset:
    """
    Data class representing the whole raw dataset
    that should be split somehow in the future.
    """

    def __init__(self, df, target_col_name: str):
        self.df = df
        self.target_col_name = target_col_name

        self._X = self.df.drop(self.target_col_name, axis=1)
        self._y = self.df[target_col_name]

    @property
    def X(self):
        return self._X

    @property
    def y(self):
        return self._y



from abc import abstractmethod
from typing import Optional

import numpy as np


class ModelInfo:
    """
    Just an interface for TravaModel to not expose its internals.
    """

    @property
    @abstractmethod
    def model_id(self) -> str:
        pass

    @abstractmethod
    def get_model(self, for_train: bool):
        pass

    @abstractmethod
    def raw_model(self):
        pass

    @abstractmethod
    def y(self, for_train: bool) -> np.array:
        pass

    @abstractmethod
    def y_pred(self, for_train: bool) -> np.array:
        pass

    @abstractmethod
    def y_pred_proba(self, for_train: bool) -> np.array:
        pass

    @property
    @abstractmethod
    def fit_time(self) -> Optional[float]:
        pass

    @property
    @abstractmethod
    def predict_time(self) -> Optional[float]:
        pass



from typing import Optional

from trava.ext.boosting_eval.boosting_logic import CommonBoostingEvalLogic


class CatBoostEvalLogic(CommonBoostingEvalLogic):
    def __init__(self, needs_plot: bool, eval_metric: str = None, early_stopping_rounds: Optional[int] = 10):
        assert not eval_metric, "Initialize your model with the eval metric instead"

        super().__init__(needs_plot=needs_plot, eval_metric=eval_metric, early_stopping_rounds=early_stopping_rounds)

    def _best_iteration(self, model) -> int:
        return model.get_best_iteration()

    def _evals_results(self, model) -> dict:
        return model.get_evals_result()

    def _train_metrics_key(self, model) -> Optional[str]:
        if self._n_eval_sets(model=model) == 1:
            return None

        return "learn"

    def _eval_metrics_key(self, model) -> str:
        return "validation"

    @property
    def _user_train_in_eval(self) -> bool:
        return False



import matplotlib.pyplot as plt
from sklearn.metrics import ConfusionMatrixDisplay, confusion_matrix

from trava.ext.results_handlers.plotter import PlotItem
from trava.ext.results_handlers.scorer_plotter import ScorerPlotter
from trava.ext.sklearn.scorers import sk
from trava.metric import Metric


class ConfMatrixPlotter(ScorerPlotter):
    def plot(self, metric: Metric, fig, ax, color: str, label: str):
        ax.set_title(f"{metric.model_id}: {label} Confusion matrix")
        conf_matrix = metric.value
        ConfusionMatrixDisplay(conf_matrix).plot(cmap=plt.cm.Blues, ax=ax)


class ConfMatrixPlotItem(PlotItem):
    def __init__(self, normalize: str, threshold=None):
        self.scorer = sk(confusion_matrix, normalize=normalize, threshold=threshold)
        self.plotter = ConfMatrixPlotter()
        self.can_overlap = False



import pytest
from copy import copy

from tests.objects_for_tests import TestSerializer
from trava.fit_predictor import FitPredictor, FitPredictorSteps
from trava.split.result import SplitResult


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


@pytest.fixture()
def raw_model(mocker, model_id):
    result = mocker.Mock()
    result.model_id = model_id
    return result


@pytest.fixture
def X_y_train(mocker):
    X_train = mocker.MagicMock()
    X_train["f1"] = [1, 2, 3]
    y_train = [0, 0, 0]

    return X_train, y_train


@pytest.fixture
def X_y_test(mocker):
    X_test = mocker.MagicMock()
    X_test["f1"] = [4, 5, 6]
    y_test = [1, 1, 1]

    return X_test, y_test


@pytest.fixture
def split_result(X_y_train, X_y_test):
    X_train, y_train = X_y_train
    X_test, y_test = X_y_test

    return SplitResult(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)


@pytest.mark.parametrize("n_steps", [1, 3])
def test_model_update_steps(mocker, n_steps, raw_model):
    config = mocker.MagicMock()
    config.fit_params = mocker.MagicMock()
    config.predict_params = mocker.MagicMock()
    config.raw_model = raw_model
    config.scorers_providers = mocker.MagicMock()

    steps = []
    for _ in range(n_steps):
        step = mocker.Mock()
        step.update_model.return_value = raw_model
        steps.append(step)

    fp_steps = FitPredictorSteps(raw_model_steps=steps)
    fp = FitPredictor(steps=fp_steps)

    fp.fit_predict(config=config, tracker=mocker.Mock())

    for step in steps:
        step.update_model.assert_called_with(raw_model=raw_model, config=config)


def test_models_configs(mocker, raw_model, model_id):
    fp = FitPredictor()
    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs = mocker.MagicMock()
    models_configs_func.return_value = models_configs

    config = mocker.Mock()
    config.raw_model = raw_model
    config.scorers_providers = mocker.MagicMock()

    fp.fit_predict(config=config, tracker=mocker.Mock())

    models_configs_func.assert_called_once_with(raw_model=raw_model, config=config)
    models_configs.__iter__.assert_called()


@pytest.mark.parametrize("n_steps", [1, 3])
def test_config_update_steps(mocker, n_steps, split_result, raw_model):
    fit_params = {"c": 2, "d": 3}
    predict_params = {"_": 4, "+": 5}

    config = mocker.Mock()
    config.raw_split_data = split_result
    config.raw_model = raw_model
    config.fit_params = fit_params
    config.predict_params = predict_params
    config.scorers_providers = mocker.MagicMock()

    steps = []
    for _ in range(n_steps):
        step = mocker.Mock()
        step.fit_split_data.return_value = split_result
        step.fit_params.return_value = fit_params
        step.predict_params.return_value = predict_params
        steps.append(step)
    fp_steps = FitPredictorSteps(config_steps=steps)
    fp = FitPredictor(steps=fp_steps)

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs_func.return_value = [(mocker.Mock(), config)]

    tracker = mocker.Mock()
    fp.fit_predict(config=config, tracker=tracker)

    for step in steps:
        step.fit_split_data.assert_called_with(raw_split_data=split_result, config=config, tracker=tracker)
        step.fit_params.assert_called_with(
            fit_params=fit_params, fit_split_data=split_result, config=config, tracker=tracker
        )
        step.predict_params.assert_called_with(
            predict_params=predict_params, fit_split_data=split_result, config=config, tracker=tracker
        )


@pytest.mark.parametrize("n_steps", [1, 3])
def test_final_steps(mocker, split_result, raw_model, n_steps):
    fit_params = {"c": 2, "d": 3}
    predict_params = {"_": 4, "+": 5}

    config = mocker.Mock()
    config.raw_split_data = split_result
    config.raw_model = raw_model
    config.fit_params = fit_params
    config.predict_params = predict_params
    config.scorers_providers = mocker.MagicMock()

    steps = []
    for _ in range(n_steps):
        step = mocker.Mock()
        step.fit_split_data.return_value = split_result
        step.fit_params.return_value = fit_params
        step.predict_params.return_value = predict_params
        steps.append(step)
    fp_steps = FitPredictorSteps(final_steps=steps)
    fp = FitPredictor(steps=fp_steps)

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    trava_model = mocker.Mock()
    models_configs_func.return_value = [(trava_model, config)]

    tracker = mocker.Mock()
    fp.fit_predict(config=config, tracker=tracker)

    for step in steps:
        step.handle.assert_called_with(trava_model=trava_model, config=config, tracker=tracker)


@pytest.mark.parametrize("n_models", [1, 3])
@pytest.mark.parametrize("has_description", [True, False], ids=["has_descr", "no_descr"])
@pytest.mark.parametrize("serializer", [TestSerializer(), None], ids=["serial", "no_serial"])
def test_tracking(mocker, model_id, split_result, raw_model, n_models, has_description, serializer):
    config = mocker.Mock()
    config.model_id = model_id
    config.raw_split_data = split_result
    config.serializer = serializer
    config.raw_model = raw_model
    model_init_params = mocker.Mock()

    description = "descr"
    if has_description:
        config.description = description
    else:
        config.description = None

    config.model_init_params = model_init_params
    fit_params = mocker.MagicMock()
    config.fit_params = fit_params
    predict_params = mocker.MagicMock()
    config.predict_params = predict_params
    config.scorers_providers = mocker.MagicMock()
    fp = FitPredictor()

    models_configs = []
    for idx in range(n_models):
        nested_model = mocker.Mock()
        nested_model.raw_model = raw_model
        nested_model_id = model_id + "_" + str(idx)
        nested_model.model_id = nested_model_id
        model_config = copy(config)
        model_config.model_id = nested_model_id
        models_configs.append((nested_model, model_config))

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs_func.return_value = models_configs

    tracker = mocker.Mock()
    fp.fit_predict(config=config, tracker=tracker)

    expected_calls = []

    is_multiple_models = n_models > 1
    if is_multiple_models:
        expected_calls += _start_tracking_check_get_calls(
            mocker=mocker,
            model_id=model_id,
            raw_model=raw_model,
            tracker=tracker,
            description=description,
            model_init_params=model_init_params,
            has_description=has_description,
            nested=False,
        )

    models = [model_config[0] for model_config in models_configs]
    for idx, model in enumerate(models):
        expected_calls += _start_tracking_check_get_calls(
            mocker=mocker,
            model_id=model.model_id,
            raw_model=raw_model,
            tracker=tracker,
            description=description,
            model_init_params=model_init_params,
            has_description=has_description,
            nested=is_multiple_models,
        )

        expected_calls += [
            mocker.call.track_fit_params(model_id=model.model_id, params=fit_params),
            mocker.call.track_predict_params(model_id=model.model_id, params=predict_params),
        ]

        model_results_kwargs = tracker.track_model_results.call_args_list[idx][1]
        model_results = model_results_kwargs["model_results"]
        assert model_results.model_id == model.model_id

        expected_calls += [mocker.call.track_model_results(model_results=model_results)]

        if serializer:
            expected_calls += [
                mocker.call.track_model_artifact(model_id=model.model_id, model=raw_model, serializer=serializer)
            ]

        expected_calls += [mocker.call.end_tracking(model_id=model.model_id)]

    if is_multiple_models:
        model_results_kwargs = tracker.track_model_results.call_args_list[n_models][1]
        model_results = model_results_kwargs["model_results"]
        assert model_results.model_id == model_id

        expected_calls += [
            mocker.call.track_model_results(model_results=model_results),
            mocker.call.end_tracking(model_id=model_id),
        ]

    tracker.assert_has_calls(expected_calls)


def _start_tracking_check_get_calls(
    mocker, model_id, raw_model, tracker, description, model_init_params, has_description: bool, nested: bool
) -> list:
    expected_calls = [
        mocker.call.start_tracking(model_id=model_id, nested=nested),
        mocker.call.track_model_init_params(model_id=model_id, params=model_init_params),
        mocker.call.track_model_info(model_id=model_id, model=raw_model),
    ]

    if has_description:
        expected_calls.append(mocker.call.track_model_description(model_id=model_id, description=description))
    else:
        tracker.track_model_description.assert_not_called()

    return expected_calls


def test_has_any_log_calls(mocker, split_result, raw_model):
    config = mocker.Mock()
    config.raw_split_data = split_result
    config.raw_model = raw_model
    config.fit_params = mocker.MagicMock()
    config.predict_params = mocker.MagicMock()
    config.scorers_providers = mocker.MagicMock()

    logger = mocker.Mock()
    fp = FitPredictor(logger=logger)
    fp.fit_predict(config=config, tracker=mocker.Mock())
    logger.log.assert_called()


@pytest.mark.parametrize("n_models", [1, 3])
def test_fit_predict(mocker, model_id, split_result, raw_model, n_models):
    config = mocker.Mock()
    config.model_id = model_id
    config.raw_split_data = split_result
    config.raw_model = raw_model
    model_init_params = mocker.Mock()

    config.model_init_params = model_init_params
    fit_params = mocker.MagicMock()
    config.fit_params = fit_params
    predict_params = mocker.MagicMock()
    config.predict_params = predict_params
    scorers_providers = mocker.MagicMock()
    config.scorers_providers = scorers_providers
    fp = FitPredictor()
    fit_mock = mocker.patch.object(fp, "_fit")
    fit_mock.return_value = None
    predict_mock = mocker.patch.object(fp, "_predict")
    predict_mock.return_value = None

    models_configs = []
    for idx in range(n_models):
        nested_model = mocker.Mock()
        nested_model.raw_model = raw_model
        nested_model_id = model_id + "_" + str(idx)
        nested_model.model_id = nested_model_id

        model_config = copy(config)

        model_config.model_id = nested_model_id
        models_configs.append((nested_model, model_config))

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs_func.return_value = models_configs

    tracker = mocker.Mock()
    fp.fit_predict(config=config, tracker=tracker)

    is_multiple_models = n_models > 1
    if is_multiple_models:
        for idx, (model, config) in enumerate(models_configs):
            fit_call_args = fit_mock.call_args_list[idx][1]
            assert fit_call_args["trava_model"] == model
            assert fit_call_args["X"] == config.raw_split_data.X_train
            assert fit_call_args["y"] == config.raw_split_data.y_train
            assert fit_call_args["fit_params"] == fit_params
            assert fit_call_args["predict_params"] == predict_params

            predict_call_args = predict_mock.call_args_list[idx][1]

            assert predict_call_args["trava_model"] == model
            assert predict_call_args["X"] == config.raw_split_data.X_test
            assert predict_call_args["y"] == config.raw_split_data.y_test
    else:
        fit_mock.assert_called_once_with(
            trava_model=models_configs[0][0],
            X=split_result.X_train,
            y=split_result.y_train,
            fit_params=fit_params,
            predict_params=predict_params,
        )

        predict_mock.assert_called_once_with(
            trava_model=models_configs[0][0], X=split_result.X_test, y=split_result.y_test
        )


@pytest.mark.parametrize("n_models", [1, 3])
def test_return_evaluators(mocker, model_id, split_result, raw_model, n_models):
    config = mocker.Mock()
    config.model_id = model_id
    config.raw_split_data = split_result
    config.raw_model = raw_model
    model_init_params = mocker.Mock()

    config.model_init_params = model_init_params
    fit_params = mocker.MagicMock()
    config.fit_params = fit_params
    predict_params = mocker.MagicMock()
    config.predict_params = predict_params
    config.scorers_providers = mocker.MagicMock()
    fp = FitPredictor()

    models_configs = []
    for idx in range(n_models):
        nested_model = mocker.Mock()
        nested_model.raw_model = raw_model
        nested_model_id = model_id + "_" + str(idx)
        nested_model.model_id = nested_model_id
        model_config = copy(config)
        model_config.model_id = nested_model_id
        models_configs.append((nested_model, model_config))

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs_func.return_value = models_configs

    tracker = mocker.Mock()
    evaluators = fp.fit_predict(config=config, tracker=tracker)

    assert len(evaluators) == n_models

    is_multiple_models = n_models > 1
    if is_multiple_models:
        models = [model_config[0] for model_config in models_configs]

        for model, evaluator in zip(models, evaluators):
            assert model.model_id == evaluator.model_id
            assert model == evaluator.trava_model
            assert split_result == evaluator.fit_split_data
            assert split_result == evaluator.raw_split_data
    else:
        model = models_configs[0][0]
        evaluator = evaluators[0]

        assert model.model_id == evaluator.model_id
        assert model == evaluator.trava_model
        assert split_result == evaluator.fit_split_data
        assert split_result == evaluator.raw_split_data


@pytest.mark.parametrize("n_models", [1, 3])
def test_evaluators_get_called(mocker, model_id, split_result, raw_model, n_models):
    config = mocker.Mock()
    config.model_id = model_id
    config.raw_split_data = split_result
    config.raw_model = raw_model
    model_init_params = mocker.Mock()

    config.model_init_params = model_init_params
    fit_params = mocker.MagicMock()
    config.fit_params = fit_params
    predict_params = mocker.MagicMock()
    config.predict_params = predict_params
    scorers_providers = mocker.MagicMock()
    config.scorers_providers = scorers_providers
    fp = FitPredictor()

    models_configs = []
    for idx in range(n_models):
        nested_model = mocker.Mock()
        nested_model.raw_model = raw_model
        nested_model_id = model_id + "_" + str(idx)
        nested_model.model_id = nested_model_id
        model_config = copy(config)
        model_config.model_id = nested_model_id
        models_configs.append((nested_model, model_config))

    models_configs_func = mocker.patch.object(fp, "_models_configs")
    models_configs_func.return_value = models_configs

    evaluator_func = mocker.patch.object(fp, "_evaluator")
    evaluator_func.return_value = mocker.Mock()

    evaluators = fp.fit_predict(config=config, tracker=mocker.Mock())

    is_multiple_models = n_models > 1
    if is_multiple_models:
        for idx in range(n_models):
            model = models_configs[idx][0]
            config = models_configs[idx][1]
            evaluator = evaluators[idx]

            eval_func_args = evaluator_func.call_args_list[idx][1]
            assert eval_func_args["model_config"] == config
            assert eval_func_args["split_result"] == config.raw_split_data
            assert eval_func_args["model"] == model

            evaluator.evaluate.assert_called_with(scorers_providers=scorers_providers)
    else:
        model = models_configs[0][0]
        config = models_configs[0][1]
        evaluator = evaluators[0]

        evaluator_func.assert_called_with(model_config=config, split_result=split_result, model=model)
        evaluator.evaluate.assert_called_with(scorers_providers=scorers_providers)



from abc import abstractmethod
from typing import List

from trava.scorer import Scorer


class ScorersProvider:
    """
    Just an interface for a source of scorers.
    """

    @property
    def provider_id(self) -> str:
        return type(self).__name__

    @property
    @abstractmethod
    def scorers(self) -> List[Scorer]:
        pass

    def metric_scorers(self) -> List[Scorer]:
        return [scorer for scorer in self.scorers if not scorer.is_other_scorer]

    def other_scorers(self) -> List[Scorer]:
        return [scorer for scorer in self.scorers if scorer.is_other_scorer]



from typing import Optional, List

from trava.logger import TravaLogger
from trava.model_serializer import ModelSerializer
from trava.scorer import Scorer
from trava.trava_tracker import TravaTracker


class LogTracker(TravaTracker):
    def __init__(self, scorers: List[Scorer], logger: TravaLogger):
        super().__init__(scorers=scorers)

        self._logger = logger

    def _start_tracking(self, model_id: str, track_name: Optional[str] = None, nested: bool = False):
        self._logger.log("Start tracking: {}".format(model_id))

    def _end_tracking(self, model_id: str):
        self._logger.log("End tracking: {}".format(model_id))

    def _track_set_tracking_group(self, group: str):
        self._logger.log("Set tracking group: {}".format(group))

    def _track_model_description(self, model_id: str, description: str):
        self._logger.log("Track description: {} - {}".format(model_id, description))

    def _track_model_init_params(self, model_id: str, params: dict):
        self._logger.log("Track init params: {} - {}".format(model_id, params))

    def _track_fit_params(self, model_id: str, params: dict):
        self._logger.log("Track fit params: {} - {}".format(model_id, params))

    def _track_predict_params(self, model_id: str, params: dict):
        self._logger.log("Track predict params: {} - {}".format(model_id, params))

    def _track_metric_value(self, model_id: str, name: str, value, step=None):
        self._logger.log("Track metric: {} - {} : {}".format(model_id, name, value))

    def _track_model_info(self, model_id: str, model):
        self._logger.log("Track model: {} - {}".format(model_id, model))

    def _track_tag(self, model_id: str, tag_key: str, tag_value):
        self._logger.log("Track tag: {} - {}".format(model_id, tag_value))

    def track_model_artifact(self, model_id: str, model, serializer: ModelSerializer):
        self._logger.log("Track artifact: {} - {}".format(model_id, model))

    def _track(self, model_id: str, *args, **kwargs):
        self._logger.log("Track something: {} - {}".format(model_id, (args, kwargs)))



from abc import ABC, abstractmethod
from typing import List

from trava.logger import TravaLogger
from trava.model_results import ModelResult
from trava.scorer import Scorer
from trava.scorers_provider import ScorersProvider
from trava.tracker import Tracker


class ResultsHandler(ABC, ScorersProvider):
    """
    Provides logic how to process and show calculated metrics to a user.
    """

    def __init__(self, scorers: List[Scorer]):
        """
        Parameters
        ----------
        scorers: list
            Scorers list we want to process and show.
        """
        self._scorers = scorers.copy()

    @property
    def scorers(self) -> List[Scorer]:
        return self._scorers

    @abstractmethod
    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        """
        Implement the logic here.

        Parameters
        ----------
        results: list of ModelResult
            Contains metrics for all the models that were trained using TraVa so far.
        logger: TravaLogger
            If you want to log something along the way
        tracker: Tracker
            If you want to tracker something along the way

        Returns
        -------
        Your handler is able to return anything you want.
        """
        pass



import pytest
import pandas as pd
import numpy as np
import random
from copy import copy

from typing import List

from trava.evaluator import Evaluator
from trava.split.result import SplitResult


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


@pytest.fixture()
def model(mocker, model_id):
    result = mocker.Mock()
    result.model_id = model_id
    return result


@pytest.fixture(scope="class")
def X_y_train():
    X_train = pd.DataFrame({"f1": [1, 2, 3]})
    y_train = np.array([0, 0, 0])

    return X_train, y_train


@pytest.fixture(scope="class")
def X_y_test():
    X_test = pd.DataFrame({"f1": [4, 5, 6]})
    y_test = np.array([1, 1, 1])

    return X_test, y_test


@pytest.fixture(scope="class")
def split_result(X_y_train, X_y_test):
    X_train, y_train = X_y_train
    X_test, y_test = X_y_test

    return SplitResult(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)


def _evaluator(model, split_result, raw_split_result):
    evaluator = Evaluator(fit_split_data=split_result, raw_split_data=raw_split_result, trava_model=model)

    return evaluator


def test_properties(model, model_id, split_result):
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)

    assert evaluator.model_id == model_id
    assert evaluator.trava_model == model
    assert evaluator.fit_split_data == split_result
    assert evaluator.raw_split_data == raw_split_result


def test_evaluate_calls_scorer(mocker, model, model_id, split_result):
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)

    results_handler = mocker.MagicMock()
    scorer = mocker.Mock()
    results_handler.metric_scorers.return_value = [scorer]
    results_handler.other_scorers.return_value = [scorer]
    results_handler.provider_id = "TestResultsHandler"
    evaluator.evaluate(scorers_providers=[results_handler])

    results_handler.assert_has_calls(
        [
            mocker.call.metric_scorers(),  # train
            mocker.call.metric_scorers(),  # test
            mocker.call.other_scorers(),  # any
        ]
    )

    def _assert(called_args, for_train: bool):
        assert called_args["trava_model"] == model
        assert called_args["for_train"] == for_train

        X = split_result.X_train if for_train else split_result.X_test
        y = split_result.y_train if for_train else split_result.y_test
        X_raw = raw_split_result.X_train if for_train else raw_split_result.X_test

        assert called_args["X"].equals(X)
        assert called_args["X_raw"].equals(X_raw)
        assert np.array_equal(called_args["y"], y)

    _assert(called_args=scorer.call_args_list[0][1], for_train=True)
    _assert(called_args=scorer.call_args_list[1][1], for_train=False)
    _assert(called_args=scorer.call_args_list[2][1], for_train=False)


def test_empty_metrics_fail(mocker, model, model_id, split_result):
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)

    provider = mocker.Mock()
    provider.provider_id = "TestProvider"
    with pytest.raises(Exception):
        evaluator.train_metrics(provider=provider)
    with pytest.raises(Exception):
        evaluator.test_metrics(provider=provider)
    with pytest.raises(Exception):
        evaluator.other_metrics(provider=provider)


@pytest.mark.parametrize("n_scorers", [1, 3])
def test_all_metrics(mocker, model, model_id, split_result, n_scorers):
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)

    results_handler = mocker.MagicMock()

    scorers = []
    train_values = []
    test_and_any_values = []
    for i in range(n_scorers):
        scorer_train_value = random.randint(0, 1000)
        scorer_test_meta_values = [random.randint(0, 1000), random.randint(0, 1000)]
        test_any_metrics_values_copy = scorer_test_meta_values.copy()

        def _build_func(train_value: int, test_meta_values: List[int]):
            def _get_metric(trava_model, for_train: bool, X, X_raw, y):
                if for_train:
                    return train_value

                # test and any metrics both use test set
                result = test_meta_values[0]
                test_meta_values.remove(result)
                return result

            return _get_metric

        scorer = mocker.MagicMock(
            side_effect=_build_func(train_value=scorer_train_value, test_meta_values=scorer_test_meta_values)
        )
        train_values.append(scorer_train_value)
        test_and_any_values.append(test_any_metrics_values_copy)
        scorers.append(scorer)

    results_handler.metric_scorers.return_value = scorers
    results_handler.other_scorers.return_value = scorers
    results_handler.provider_id = "TestResultsHandler"
    evaluator.evaluate(scorers_providers=[results_handler])

    train_metrics = evaluator.train_metrics(provider=results_handler)
    test_metrics = evaluator.test_metrics(provider=results_handler)
    any_metrics = evaluator.other_metrics(provider=results_handler)

    for idx, metric in enumerate(train_metrics):
        assert metric.value == train_values[idx]

    for idx, metric in enumerate(test_metrics):
        assert metric.value == test_and_any_values[idx][0]

    for idx, metric in enumerate(any_metrics):
        assert metric.value == test_and_any_values[idx][1]


def test_unload_data():
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)
    evaluator.unload_data()

    assert not evaluator.fit_split_data
    assert not evaluator.raw_split_data


def test_metrics_caching(mocker, model, model_id, split_result):
    raw_split_result = copy(split_result)
    evaluator = _evaluator(model=model, split_result=split_result, raw_split_result=raw_split_result)

    scorer_name = "test_scorer_name"
    scorers_provider = mocker.MagicMock()

    scorer_1 = mocker.Mock()
    scorer_1.func_name = scorer_name

    scorer_2 = mocker.Mock()
    scorer_2.func_name = scorer_name

    scorer_3 = mocker.Mock()
    scorer_3.func_name = "another_scorer"

    scorers = [
        scorer_1,
        scorer_2,
        scorer_3,
    ]

    scorers_provider.metric_scorers.return_value = scorers
    scorers_provider.other_scorers.return_value = []

    evaluator.evaluate(scorers_providers=[scorers_provider])

    scorer_1.assert_called()
    scorer_2.assert_not_called()
    scorer_3.assert_called()



import numbers

from trava.utils.model_params_filter import merge_given_params_with_default, filter_params, _allowed_params_types


def test_merge_given_params():
    class TestObject:
        def __init__(self, a: bool, b: str = "preved", c: int = 10, d: tuple = ()):
            pass

    params = {"a": True, "c": 999}

    result = merge_given_params_with_default(object_type=TestObject, params=params)

    true_result = {"a": True, "b": "preved", "c": 999}

    assert result == true_result


def test_filter_params():
    params = {"a": 1, "b": "b", "c": [1, 2, 3], "d": Exception, "e": test_filter_params, "f": 123.0}

    result = filter_params(params=params)

    true_params = {"a": 1, "b": "b", "f": 123.0}

    assert result == true_params


def test_allowed_init_params_types():
    test_init_params_types = (
        numbers.Number,
        str,
        bool,
    )
    assert _allowed_params_types() == test_init_params_types



from pickle import load, dump  # type: ignore
from trava.model_serializer import ModelSerializer


class PickleModelSerializer(ModelSerializer):
    def load(self, path: str):
        with open(path, "r") as file:
            return load(file)

    def save(self, model, path: str):
        with open(path, "wb") as file:
            return dump(model, file)



import pandas as pd

from trava.ext.split_logic.basic import BasicSplitLogic
from trava.split.logic import SplitLogic


class WithinGroupSplitLogic(SplitLogic):
    """
    Performs a split for every group in the data and then
    merges them into the single result.
    Assumes that you work with pandas data frame.
    Init parameters

    ----------
    group_col:
        What is column that contains group identifiers.
    group_split_logic: SplitLogic
        Contains logic how to split each group.
    group_split_logic_kwargs: dict
        Any additional parameters you want to pass to group_split_logic.
    """

    def __init__(self, group_col: str, group_split_logic: SplitLogic, group_split_logic_kwargs: dict = None):
        self._group_col = group_col
        self._group_split_logic = group_split_logic or BasicSplitLogic()
        self._group_split_logic_kwargs = group_split_logic_kwargs if group_split_logic_kwargs else {}

    def split(self, data, test_size: float, valid_size: float, **kwargs) -> tuple:
        unique_groups = data[self._group_col].unique()

        all_groups_data = []

        for group in unique_groups:
            group_data = data[data[self._group_col] == group]

            split_group_data = self._group_split_logic.split(
                group_data, test_size=test_size, valid_size=valid_size, **self._group_split_logic_kwargs
            )
            if not valid_size:
                # to safely merge dataframes after split within each group
                split_group_data = split_group_data[:2]
            all_groups_data.append(split_group_data)

        separated_groups_data = list(zip(*all_groups_data))
        merged_groups_data = [pd.concat(dfs, axis=0, copy=False, sort=False) for dfs in separated_groups_data]
        if not valid_size:
            # to conform to expected output
            merged_groups_data.append(None)
        return tuple(merged_groups_data)



import pytest
import numpy as np
import pandas as pd

from tests.objects_for_tests import TestResultsHandler, TestScorer, TestAnyScorer, TestModel
from trava.evaluator import Evaluator
from trava.fit_predictor import FitPredictConfig
from trava.metric import Metric
from trava.raw_dataset import RawDataset
from trava.split.result import SplitResult
from trava.trava_tracker import TravaTracker
from trava.trava_sv import TravaSV


@pytest.fixture(scope="class")
def trava_tracker():
    return TravaTracker(scorers=[])


@pytest.fixture(scope="class")
def test_result_handlers():
    return [TestResultsHandler(scorers=[TestScorer(score_func=np.sum)])]


@pytest.fixture
def trava(test_result_handlers, trava_tracker):
    return TravaSV(results_handlers=test_result_handlers, tracker=trava_tracker)


@pytest.fixture(scope="class")
def split_result(X_y_train, X_y_test):
    X_train, y_train = X_y_train
    X_test, y_test = X_y_test

    return SplitResult(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)


@pytest.fixture(scope="class")
def X_y_train():
    X_train = pd.DataFrame({"f1": [1, 2, 3]})
    y_train = np.array([0, 0, 0])

    return X_train, y_train


@pytest.fixture(scope="class")
def X_y_test():
    X_test = pd.DataFrame({"f1": [4, 5, 6]})
    y_test = np.array([1, 1, 1])

    return X_test, y_test


@pytest.fixture(scope="class")
def raw_dataset():
    result = RawDataset(df=pd.DataFrame({"df": [1, 2, 3], "test_target": [0, 0, 1]}), target_col_name="test_target")
    return result


@pytest.fixture(scope="class")
def model_init_params():
    model_init_params = {
        "test1": 1,
        "test2": "second",
    }
    return model_init_params


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


@pytest.fixture(scope="class")
def main_model_id(model_id):
    return "main_" + model_id


@pytest.fixture()
def raw_model(mocker):
    model_mock = mocker.Mock()
    return model_mock


@pytest.fixture()
def trava_model(mocker, raw_model):
    model_mock = mocker.Mock()
    trava_model.raw_model = raw_model
    return model_mock


@pytest.fixture
def model_type(mocker, raw_model):
    model_type = mocker.stub()
    model_type.return_value = raw_model

    return model_type


def _fit_predictor(mocker, evaluators):
    fit_predictor = mocker.Mock()
    fit_predictor.fit_predict.return_value = evaluators
    return fit_predictor


def _mock_evaluator(
    mocker, model_id, train_values=None, test_values=None, any_values=None, trava_model=None, split_result=None
):
    evaluator = mocker.Mock()
    evaluator.model_id = model_id

    if train_values:
        train_metrics = [_metric(value=value, model_id=model_id) for value in train_values]
        evaluator.train_metrics.return_value = train_metrics

    if test_values:
        test_metrics = [_metric(value=value, model_id=model_id) for value in test_values]
        evaluator.test_metrics.return_value = test_metrics

    if any_values:
        any_metrics = [_metric(value=value, model_id=model_id) for value in any_values]
        evaluator.other_metrics.return_value = any_metrics

    if trava_model:
        evaluator.trava_model = trava_model

    if split_result:
        evaluator.fit_split_data = split_result

    if raw_dataset:
        evaluator.raw_split_data = split_result

    return evaluator


def _true_evaluator(model, split_result):
    return Evaluator(trava_model=model, fit_split_data=split_result, raw_split_data=split_result)


def _metric(model_id, value):
    return Metric(name=model_id + "_" + str(value), value=value, model_id=model_id)


def _fit_with_n_evaluators(
    mocker,
    trava_model,
    model_id,
    main_model_id,
    model_type,
    split_result,
    trava,
    n_evaluators,
    test_result_handlers,
    only_calculate_metrics: bool = False,
    keep_data_in_memory: bool = True,
    mock_evaluators: bool = True,
):
    evaluators = []
    for i in range(n_evaluators):
        indexed_model_id = model_id + "_" + str(i)
        eval_model_id = indexed_model_id if n_evaluators > 1 else model_id
        if mock_evaluators:
            evaluator = _mock_evaluator(
                mocker,
                model_id=eval_model_id,
                train_values=mocker.MagicMock(),
                test_values=mocker.MagicMock(),
                any_values=mocker.MagicMock(),
                trava_model=trava_model,
                split_result=split_result,
            )
        else:
            trava_model.model_id = eval_model_id
            evaluator = _true_evaluator(model=trava_model, split_result=split_result)
            # since fit predictor is mocked I need to evaluate the true evaluator manually
            evaluator.evaluate(scorers_providers=test_result_handlers)
        evaluators.append(evaluator)

    fit_predictor = _fit_predictor(mocker=mocker, evaluators=evaluators)
    trava.fit_predict(
        raw_split_data=split_result,
        model_id=main_model_id if n_evaluators > 1 else model_id,
        model_type=model_type,
        fit_predictor=fit_predictor,
        only_calculate_metrics=only_calculate_metrics,
        keep_data_in_memory=keep_data_in_memory,
    )
    return evaluators


# TESTS


def test_model_init(trava: TravaSV, split_result: SplitResult, model_init_params, model_type, model_id):
    trava.fit_predict(
        raw_split_data=split_result, model_id=model_id, model_type=model_type, model_init_params=model_init_params
    )

    model_type.assert_called_once_with(**model_init_params)


def test_fit_predictor(
    mocker,
    raw_model,
    trava: TravaSV,
    raw_dataset: RawDataset,
    split_result: SplitResult,
    model_init_params,
    model_type,
    model_id,
    test_result_handlers,
    trava_tracker,
):
    fit_predictor = _fit_predictor(mocker=mocker, evaluators=[])

    keep_models_in_memory = True
    serializer = mocker.Mock()
    fit_params = {"a": 1, "b": 2}
    predict_params = {"c": 3, "d": 4}
    description = "test_description"
    trava.fit_predict(
        raw_split_data=split_result,
        model_id=model_id,
        model_type=model_type,
        description=description,
        model_init_params=model_init_params,
        fit_predictor=fit_predictor,
        fit_params=fit_params,
        predict_params=predict_params,
        keep_models_in_memory=keep_models_in_memory,
        serializer=serializer,
    )

    fit_predictor.fit_predict.assert_called_once()

    test_fit_config = FitPredictConfig(
        raw_split_data=split_result,
        raw_model=raw_model,
        model_init_params=model_init_params,
        model_id=model_id,
        scorers_providers=test_result_handlers + [trava_tracker],
        serializer=serializer,
        fit_params=fit_params,
        predict_params=predict_params,
        description=description,
    )

    fit_predictor.fit_predict.assert_called_once_with(config=test_fit_config, tracker=trava_tracker)


@pytest.mark.parametrize(
    "evaluator_values",
    [
        [
            ([1.0, 3.0, {}, "train_string"], [4.0, 4.0, [], "test_string"], [10, pd.DataFrame, []])  # train  # test
        ],  # model meta
        [
            (
                [1.0, 3.0, {}, "train_string"],  # train
                [4.0, 4.0, [], "test_string"],  # test
                [10, pd.DataFrame, []],
            ),  # model meta
            (
                [9.0, 5.0, {}, "train_string"],  # train
                [16.0, 8.0, [], "test_string"],  # test
                [20, pd.DataFrame, []],
            ),  # model meta
        ],
    ],
)
def test_fit_predict_many_result(
    mocker, trava: TravaSV, split_result: SplitResult, model_type, model_id, evaluator_values
):
    evaluators = []

    for values in evaluator_values:
        evaluator = _mock_evaluator(
            mocker, model_id=model_id, train_values=values[0], test_values=values[1], any_values=values[2]
        )

        evaluators.append(evaluator)

    fit_predictor = _fit_predictor(mocker=mocker, evaluators=evaluators)

    result = trava.fit_predict(
        raw_split_data=split_result, model_id=model_id, model_type=model_type, fit_predictor=fit_predictor
    )
    if len(evaluator_values) == 1:
        assert result == [[evaluator_values[0][0], evaluator_values[0][1], evaluator_values[0][2]]]
    else:
        assert result == [[[5.0, 4.0], [10.0, 6.0], [15.0]]]

    assert trava.detailed_results_for(model_id=model_id)


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_no_save(
    mocker, trava, model_id, main_model_id, model_type, trava_model, split_result, test_result_handlers, n_evaluators
):
    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=test_result_handlers,
    )

    trava.evaluate(
        model_id=main_model_id if n_evaluators > 1 else model_id,
        results_handlers=test_result_handlers,
        save_results=False,
    )

    for evaluator in evaluators:
        evaluator.evaluate.assert_not_called()


# noinspection PyPep8Naming
@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_save(
    mocker,
    trava,
    model_id,
    main_model_id,
    raw_model,
    trava_model,
    model_type,
    split_result,
    X_y_train,
    X_y_test,
    n_evaluators,
):
    any_and_test_data = 1233
    trava_model.get_model.return_value = raw_model
    trava_model.any_and_test_data = any_and_test_data

    result_handlers = [
        TestResultsHandler(
            scorers=[
                TestScorer(score_func=np.sum),
                TestAnyScorer(model_func=lambda scorer_model: scorer_model.any_and_test_data),
            ]
        )
    ]

    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=result_handlers,
    )

    result = trava.evaluate(
        model_id=main_model_id if n_evaluators > 1 else model_id, results_handlers=result_handlers, save_results=True
    )

    for evaluator in evaluators:
        evaluator.evaluate.assert_called_once_with(scorers_providers=result_handlers)

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    f1_train_sum = np.sum(X_train["f1"])
    f1_test_sum = np.sum(X_test["f1"])

    assert result == [[[f1_train_sum], [f1_test_sum], [any_and_test_data]]]


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_track(
    mocker,
    trava,
    model_id,
    main_model_id,
    model_type,
    trava_model,
    split_result,
    test_result_handlers,
    X_y_train,
    X_y_test,
    n_evaluators,
):
    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=test_result_handlers,
    )

    tracker_mock = mocker.Mock()
    tracker_copy_func = mocker.patch.object(trava, "_tracker_copy")
    tracker_copy_func.return_value = tracker_mock

    scorer_1 = mocker.Mock()
    scorer_2 = mocker.Mock()

    scorers = [scorer_1, scorer_2]

    trava.evaluate_track(model_id=main_model_id if n_evaluators > 1 else model_id, scorers=scorers)
    calls = [mocker.call.add_scorers(scorers=scorers)]
    for idx, evaluator in enumerate(evaluators):
        calls += [
            mocker.call.start_tracking(model_id=evaluator.model_id),
            mocker.call.track_model_results(model_results=mocker.ANY),
            mocker.call.end_tracking(model_id=evaluator.model_id),
        ]

        track_model_results_kwargs = tracker_mock.track_model_results.call_args_list[idx][1]
        model_results = track_model_results_kwargs["model_results"]

        assert (
            model_results.model_id == evaluator.model_id
            and len(model_results.evaluators) == 1
            and model_results.evaluators == [evaluator]
        )

    if n_evaluators > 1:
        calls += [
            mocker.call.start_tracking(model_id=main_model_id),
            mocker.call.track_model_results(model_results=mocker.ANY),
            mocker.call.end_tracking(model_id=main_model_id),
        ]

        track_model_results_kwargs = tracker_mock.track_model_results.call_args_list[n_evaluators][1]
        model_results = track_model_results_kwargs["model_results"]

        assert (
            model_results.model_id == main_model_id
            and len(model_results.evaluators) == n_evaluators
            and model_results.evaluators == evaluators
        )

    tracker_mock.assert_has_calls(calls)


def test_detailed_results_no_model(trava):
    with pytest.raises(AttributeError):
        trava.detailed_results_for(model_id="arbitrary_id")


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_detailed_results(
    mocker,
    trava,
    model_id,
    main_model_id,
    model_type,
    trava_model,
    split_result,
    test_result_handlers,
    X_y_train,
    X_y_test,
    n_evaluators,
):
    test_model_id = main_model_id if n_evaluators > 1 else model_id

    meta_test_data = 1233
    trava_model.meta_test_data = meta_test_data

    result_handler = TestResultsHandler(
        scorers=[
            TestScorer(score_func=np.sum),
            TestAnyScorer(model_func=lambda scorer_model: scorer_model.meta_test_data),
        ]
    )
    _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=[result_handler],
        mock_evaluators=False,
    )

    result = trava.detailed_results_for(model_id=test_model_id)

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    f1_train_sum = np.sum(X_train["f1"])
    f1_test_sum = np.sum(X_test["f1"])

    assert result == [[[f1_train_sum] * n_evaluators, [f1_test_sum] * n_evaluators, [meta_test_data] * n_evaluators]]


# TODO: move to trava_base tests
@pytest.mark.parametrize("n_fits", [1, 2])
def test_results(
    mocker,
    trava,
    model_id,
    main_model_id,
    model_type,
    trava_model,
    split_result,
    test_result_handlers,
    X_y_train,
    X_y_test,
    n_fits,
):
    score_funcs = [np.sum, np.mean]

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    meta_test_data = 1233
    trava_model.meta_test_data = meta_test_data

    metrics_values = []
    for i in range(n_fits):
        score_func = score_funcs[i]
        result_handler = TestResultsHandler(
            scorers=[
                TestScorer(score_func=score_func),
                TestAnyScorer(model_func=lambda scorer_model: scorer_model.meta_test_data),
            ]
        )
        _fit_with_n_evaluators(
            mocker=mocker,
            trava_model=trava_model,
            model_id=model_id + "_" + str(i),
            main_model_id=main_model_id,
            model_type=model_type,
            split_result=split_result,
            trava=trava,
            n_evaluators=1,
            test_result_handlers=[result_handler],
            mock_evaluators=False,
        )

        metrics_values.append((score_func(X_train["f1"]), score_func(X_test["f1"]), meta_test_data))

    results = trava.results

    test_metrics_values = [list(tup) for tup in list(zip(*metrics_values))]
    # test_metrics_values.append([])
    assert results == [test_metrics_values]


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_dont_keep_data(
    mocker, trava, model_id, main_model_id, model_type, trava_model, split_result, test_result_handlers, n_evaluators
):
    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=test_result_handlers,
        keep_data_in_memory=True,
    )

    for evaluator in evaluators:
        evaluator.unload_data.assert_not_called()

    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=test_result_handlers,
        keep_data_in_memory=False,
    )

    for evaluator in evaluators:
        evaluator.unload_data.assert_called()


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_raw_models(
    mocker,
    trava,
    model_id,
    main_model_id,
    model_type,
    raw_model,
    trava_model,
    split_result,
    test_result_handlers,
    n_evaluators,
):
    evaluators = _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        test_result_handlers=test_result_handlers,
        keep_data_in_memory=False,
        mock_evaluators=False,
    )

    target_model_id = model_id if n_evaluators == 1 else main_model_id
    raw_models = trava.raw_models_for(model_id=target_model_id)

    test_raw_models = dict([(evaluator.model_id, evaluator.trava_model.raw_model) for evaluator in evaluators])

    assert raw_models == test_raw_models


def test_create_raw_model(mocker, trava):
    model_type = TestModel
    def_1_value = "custom_value"
    req_param_value = 111
    model_init_params = {"required_param_1": mocker.Mock(), "required_param_2": req_param_value, "def_1": def_1_value}
    model, result_init_params = trava._create_raw_model(model_type=model_type, model_init_params=model_init_params)

    test_init_params = {"required_param_2": req_param_value, "def_1": def_1_value, "def_2": "tada", "def_5": False}

    assert isinstance(model, model_type)
    assert result_init_params == test_init_params


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_only_calculate_metrics(mocker, model_id, main_model_id, model_type, trava_model, split_result, n_evaluators):
    results_handler = mocker.Mock()
    results_handlers = [results_handler]
    trava = TravaSV(results_handlers=results_handlers, tracker=trava_tracker)
    _fit_with_n_evaluators(
        mocker=mocker,
        trava_model=trava_model,
        model_id=model_id,
        main_model_id=main_model_id,
        model_type=model_type,
        split_result=split_result,
        trava=trava,
        n_evaluators=n_evaluators,
        only_calculate_metrics=True,
        test_result_handlers=results_handlers,
        keep_data_in_memory=True,
    )

    results_handler.handle.assert_not_called()



from typing import List

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.scorer import Scorer
from trava.trava_tracker import Tracker


class MetricsDictHandler(ResultsHandler):
    """
    Returns all the metrics wrapped in a dictionary.
    """

    def __init__(self, scorers: List[Scorer], include_train_metrics: bool = False):
        super().__init__(scorers)

        self._include_train_metrics = include_train_metrics

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        result = {}

        for model_results in results:
            model_metrics = {}
            if self._include_train_metrics:
                model_metrics["train"] = self._metrics_dict(metrics=model_results.train_metrics(provider=self))

            model_metrics["test"] = self._metrics_dict(metrics=model_results.test_metrics(provider=self))
            any_metrics = model_results.other_metrics(provider=self)
            model_metrics["other"] = self._metrics_dict(metrics=any_metrics)

            result[model_results.model_id] = model_metrics

        return result

    @staticmethod
    def _metrics_dict(metrics: List[Metric]):
        result = {}

        for metric in metrics:
            if metric.is_scalar:
                result[metric.name] = metric.value

        return result



from abc import ABC, abstractmethod


class SplitLogic(ABC):
    """
    Encapsulates logic of how to split data.
    """

    @abstractmethod
    def split(self, data, test_size: float, valid_size: float, **kwargs) -> tuple:
        """
        Splits the data

        Parameters
        ----------
        data: any data type your split logic supports
        test_size: float
            fraction of data to use as test set.
        valid_size: float
            fraction of data to use as validation set. Pass 0.0 if you don't want validation set.
        kwargs: dict
            Any additional parameters your class works with. Note: Init injection is more preferable.

        Returns
        -------
        Tuple of (train, test, validation) data
        """
        pass



from typing import List, Optional, Tuple, Iterator
import numpy as np

from trava.evaluator import Evaluator
from trava.metric import Metric
from trava.scorers_provider import ScorersProvider


class ModelResult:
    """
    Helps to get metrics for a model based on its evaluators.
    There can be more than one fit per model when you
    train it on different sets of data ( e.g. using CV ).
    Model results contains an evaluator for each fit
    and helps to aggregate metrics from multiple fits.

    Init parameters
    ----------
    model_id: str
        Model unique identifier, will be used for saving metrics etc
    evaluators:
        All the fits that were made with the same model.
    """

    def __init__(self, model_id: str, evaluators: List[Evaluator]):
        assert len(evaluators) > 0, "At least one evaluator is required"

        self._model_id = model_id
        self._evaluators = evaluators

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def is_one_fit_result(self) -> bool:
        return len(self._evaluators) == 1

    @property
    def raw_models(self) -> Optional[dict]:
        """
        Gets the raw models

        Returns
        -------
        List of trained models
        """
        return dict([(evaluator.model_id, evaluator.trava_model.raw_model) for evaluator in self._evaluators])

    @property
    def evaluators(self) -> List[Evaluator]:
        """
        Gets all the evaluators provided

        Returns
        -------
        List of the trained models
        """
        return self._evaluators

    def train_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Parameters
        ----------
        provider: ScorersProvider
            Provides scorers to calculate metrics

        Returns
        -------
        All metrics calculated on the train set.
        """
        return self._metrics(provider=provider, on="train")

    def test_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Parameters
        ----------
        provider: ScorersProvider
            Provides scorers to calculate metrics

        Returns
        -------
        All metrics calculated on the test set.
        """
        return self._metrics(provider=provider, on="test")

    def other_metrics(self, provider: ScorersProvider) -> List[Metric]:
        """
        Parameters
        ----------
        provider: ScorersProvider
            Provides scorers to calculate metrics

        Returns
        -------
        All other metrics calculated on the test set.
        """
        return self._metrics(provider=provider, on="any")

    def _metrics(self, provider: ScorersProvider, on: str) -> List[Metric]:
        all_metrics = [
            self._get_metrics(provider=provider, evaluator=evaluator, on=on) for evaluator in self._evaluators
        ]
        return self._calculate_metrics(metrics=all_metrics)

    def _calculate_metrics(self, metrics: List[List[Metric]]):
        """
        If there is only one evaluator, returns all metrics as it is.
        If there are many evaluators, than it averages out all scalar metrics
        and ignores all the complex metrics.
        """
        first_evaluator_metrics = metrics[0]

        if first_evaluator_metrics is None:
            return None

        if len(self._evaluators) == 1:
            return first_evaluator_metrics

        result = []

        zipped_all_metrics: Iterator[Tuple[Metric, ...]] = zip(*metrics)
        for zipped_metrics in zipped_all_metrics:
            first_metric: Metric = zipped_metrics[0]

            if not first_metric.is_scalar:
                continue

            mean_value = np.mean([metric.value for metric in zipped_metrics])

            mean_metric = type(first_metric)(name=first_metric.name, value=mean_value, model_id=self._model_id)

            result.append(mean_metric)

        return result

    @staticmethod
    def _get_metrics(provider: ScorersProvider, evaluator: Evaluator, on: str) -> List[Metric]:
        if on == "train":
            return evaluator.train_metrics(provider=provider)
        elif on == "test":
            return evaluator.test_metrics(provider=provider)
        elif on == "any":
            return evaluator.other_metrics(provider=provider)
        else:
            raise ValueError("Unknown source of metrics {}".format(on))



import pytest
import numpy as np
import random

from trava.metric import Metric
from trava.model_results import ModelResult


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


def test_evaluators(mocker, model_id):
    evaluators = [mocker.Mock(), mocker.Mock()]
    model_result = ModelResult(model_id=model_id, evaluators=evaluators)
    assert model_result.evaluators == evaluators


def test_model_id(mocker, model_id):
    evaluators = [mocker.Mock()]
    model_result = ModelResult(model_id=model_id, evaluators=evaluators)
    assert model_result.model_id == model_id


def test_evaluators_required(mocker, model_id):
    evaluators = []
    with pytest.raises(Exception):
        ModelResult(model_id=model_id, evaluators=evaluators)
    with pytest.raises(Exception):
        ModelResult(model_id=model_id, evaluators=None)


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_is_one_fit_result(mocker, model_id, n_evaluators):
    evaluators = [mocker.Mock() for _ in range(n_evaluators)]
    model_result = ModelResult(model_id=model_id, evaluators=evaluators)

    if n_evaluators > 1:
        assert not model_result.is_one_fit_result
    else:
        assert model_result.is_one_fit_result


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_models(mocker, model_id, n_evaluators):
    raw_models = {}
    evaluators = []
    for idx in range(n_evaluators):
        evaluator = mocker.Mock()
        trava_model = mocker.Mock()
        raw_model = mocker.Mock()
        trava_model.raw_model = raw_model
        evaluator.model_id = model_id + "_" + str(idx)
        evaluator.trava_model = trava_model

        evaluators.append(evaluator)
        raw_models[evaluator.model_id] = raw_model

    model_result = ModelResult(model_id=model_id, evaluators=evaluators)

    assert model_result.raw_models == raw_models


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_train_metrics(mocker, n_evaluators):
    scorers_provider = mocker.Mock()

    main_model_id = "main_model_id"
    all_metrics = []
    evaluators = []
    for idx in range(n_evaluators):
        evaluator = mocker.Mock()
        if n_evaluators > 1:
            evaluator_model_id = "model_" + str(idx)
        else:
            evaluator_model_id = main_model_id

        evaluator.model_id = evaluator_model_id
        metrics = [
            Metric(name="test_metric_1", value=random.randint(0, 1000), model_id=evaluator_model_id),
            Metric(name="test_metric_2", value="blahblah", model_id=evaluator_model_id),
        ]
        all_metrics.append(metrics)

        def _wrapper(metrics):
            def _train_metrics(provider):
                if provider != scorers_provider:
                    raise ValueError
                return metrics.copy()

            return _train_metrics

        evaluator.train_metrics = mocker.MagicMock(side_effect=_wrapper(metrics=metrics))
        evaluators.append(evaluator)

    model_result = ModelResult(model_id=model_id, evaluators=evaluators)

    if n_evaluators > 1:
        filtered_all_metrics = []
        for i in range(len(all_metrics)):
            filtered_all_metrics.append([metric for metric in all_metrics[i] if metric.is_scalar])
        all_metrics = filtered_all_metrics

    for metric, eval_metrics in zip(model_result.train_metrics(provider=scorers_provider), zip(*all_metrics)):
        first_eval_metric = eval_metrics[0]

        assert metric.name == first_eval_metric.name
        if n_evaluators > 1:
            if not metric.is_scalar:
                continue
            eval_metrics_value = np.mean([metric.value for metric in eval_metrics])
        else:
            eval_metrics_value = metric.value

        assert metric.value == eval_metrics_value


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_test_metrics(mocker, n_evaluators):
    scorers_provider = mocker.Mock()

    main_model_id = "main_model_id"
    all_metrics = []
    evaluators = []
    for idx in range(n_evaluators):
        evaluator = mocker.Mock()
        if n_evaluators > 1:
            evaluator_model_id = "model_" + str(idx)
        else:
            evaluator_model_id = main_model_id

        evaluator.model_id = evaluator_model_id
        metrics = [
            Metric(name="test_metric_1", value=random.randint(0, 1000), model_id=evaluator_model_id),
            Metric(name="test_metric_2", value="blahblah", model_id=evaluator_model_id),
        ]
        all_metrics.append(metrics)

        def _wrapper(metrics):
            def _train_metrics(provider):
                if provider != scorers_provider:
                    raise ValueError
                return metrics.copy()

            return _train_metrics

        evaluator.test_metrics = mocker.MagicMock(side_effect=_wrapper(metrics=metrics))
        evaluators.append(evaluator)

    model_result = ModelResult(model_id=model_id, evaluators=evaluators)

    if n_evaluators > 1:
        filtered_all_metrics = []
        for i in range(len(all_metrics)):
            filtered_all_metrics.append([metric for metric in all_metrics[i] if metric.is_scalar])
        all_metrics = filtered_all_metrics

    for metric, eval_metrics in zip(model_result.test_metrics(provider=scorers_provider), zip(*all_metrics)):
        first_eval_metric = eval_metrics[0]

        assert metric.name == first_eval_metric.name
        if n_evaluators > 1:
            if not metric.is_scalar:
                continue
            eval_metrics_value = np.mean([metric.value for metric in eval_metrics])
        else:
            eval_metrics_value = metric.value

        assert metric.value == eval_metrics_value


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_any_metrics(mocker, n_evaluators):
    scorers_provider = mocker.Mock()

    main_model_id = "main_model_id"
    all_metrics = []
    evaluators = []
    for idx in range(n_evaluators):
        evaluator = mocker.Mock()
        if n_evaluators > 1:
            evaluator_model_id = "model_" + str(idx)
        else:
            evaluator_model_id = main_model_id

        evaluator.model_id = evaluator_model_id
        metrics = [
            Metric(name="test_metric_1", value=random.randint(0, 1000), model_id=evaluator_model_id),
            Metric(name="test_metric_2", value="blahblah", model_id=evaluator_model_id),
        ]
        all_metrics.append(metrics)

        def _wrapper(metrics):
            def _train_metrics(provider):
                if provider != scorers_provider:
                    raise ValueError
                return metrics.copy()

            return _train_metrics

        evaluator.other_metrics = mocker.MagicMock(side_effect=_wrapper(metrics=metrics))
        evaluators.append(evaluator)

    model_result = ModelResult(model_id=model_id, evaluators=evaluators)

    if n_evaluators > 1:
        filtered_all_metrics = []
        for i in range(len(all_metrics)):
            filtered_all_metrics.append([metric for metric in all_metrics[i] if metric.is_scalar])
        all_metrics = filtered_all_metrics

    for metric, eval_metrics in zip(model_result.other_metrics(provider=scorers_provider), zip(*all_metrics)):
        first_eval_metric = eval_metrics[0]

        assert metric.name == first_eval_metric.name
        if n_evaluators > 1:
            if not metric.is_scalar:
                continue
            eval_metrics_value = np.mean([metric.value for metric in eval_metrics])
        else:
            eval_metrics_value = metric.value

        assert metric.value == eval_metrics_value



from abc import ABC, abstractmethod
from collections import Counter
from typing import List, Tuple, Optional

from trava.evaluator import Evaluator
from trava.logger import TravaLogger
from trava.model_serializer import ModelSerializer
from trava.trava_model import TravaModel
from trava.model_results import ModelResult
from trava.scorers_provider import ScorersProvider
from trava.split.result import SplitResult
from trava.tracker import Tracker


class FitPredictConfig:
    """
    Data class, containing all the
    information needed for fit&predict process.
    """

    def __init__(
        self,
        raw_model,
        model_init_params: Optional[dict],
        model_id: str,
        scorers_providers: List[ScorersProvider],
        serializer: Optional[ModelSerializer],
        raw_split_data: Optional[SplitResult] = None,
        fit_params: dict = None,
        predict_params: dict = None,
        description: Optional[str] = None,
    ):
        self.raw_model = raw_model
        self.model_init_params = model_init_params or {}
        self.model_id = model_id
        self.scorers_providers = scorers_providers
        self.serializer = serializer
        self.raw_split_data = raw_split_data
        self.fit_params = fit_params or {}
        self.predict_params = predict_params or {}
        self.description = description

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FitPredictConfig):
            return False

        model_check = self.raw_model == other.raw_model
        init_params_check = self.model_init_params == other.model_init_params
        model_id_check = self.model_id == other.model_id
        scorers_providers_check = self.scorers_providers == other.scorers_providers
        serializer_check = self.serializer == other.serializer
        raw_split_data_check = self.raw_split_data == other.raw_split_data
        fit_params_check = self.fit_params == other.fit_params
        predict_params_check = self.predict_params == other.predict_params
        description_check = self.description == other.description

        result = (
            model_check
            and init_params_check
            and model_id_check
            and scorers_providers_check
            and serializer_check
            and raw_split_data_check
            and fit_params_check
            and predict_params_check
            and description_check
        )  # noqa: E127

        return result


class RawModelUpdateStep(ABC):
    """
    Using this step you can update properties of a raw model object.
    """

    @abstractmethod
    def update_model(self, raw_model, config: FitPredictConfig):
        """
        Updates the given raw model.

        Parameters
        ----------
        raw_model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.
        config: FitPredictConfig
            Initial config object

        Returns
        -------
        Updated raw model
        """
        pass


class FitPredictConfigUpdateStep:
    """
    Using this step you can update parameters containing
    in FitPredictConfig to then use them in fit&predict process.
    All methods will be called in the presented order.
    """

    def fit_split_data(self, raw_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker) -> SplitResult:
        """
        Prepares split data to fit a model with.

        Parameters
        ----------
        raw_split_data: SplitResult
            Current split_data state before applying any steps or just after another step
        config: FitPredictConfig:
            Model's config object
        tracker: Tracker
            Tracker object

        Returns
        -------
        Updated split data
        """
        return raw_split_data

    def fit_params(
        self, fit_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        """
        Prepares fit parameters to be used with a model.

        Parameters
        ----------
        fit_params: dict
            Current fit_params state before applying any steps or just after another step.
        fit_split_data: SplitResult
            Final split data that will be used to fit the model.
        config: FitPredictConfig
            Model's config object
        tracker: Tracker
            Tracker object

        Returns
        -------
        Updated fit parameters
        """
        return fit_params

    def predict_params(
        self, predict_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        """
        Prepares predict parameters to be used with a model.

        Parameters
        ----------
        predict_params: dict
            Current predict_params state before applying any steps or just after another step.
        fit_split_data: SplitResult
            Final split data that will be used to fit the model.
        config: FitPredictConfig
            Model's config object
        tracker: Tracker
            Tracker object

        Returns
        -------
        Updated predict parameters
        """
        return predict_params


class FinalHandlerStep(ABC):
    """
    Using this step you can do whatever you want after fit&predict process
    using a resulting TravaModel object.
    """

    @abstractmethod
    def handle(self, trava_model: TravaModel, config: FitPredictConfig, tracker: Tracker):
        pass


class FitPredictorSteps:
    def __init__(
        self,
        raw_model_steps: List[RawModelUpdateStep] = None,
        config_steps: List[FitPredictConfigUpdateStep] = None,
        final_steps: List[FinalHandlerStep] = None,
    ):
        self.raw_model_steps = raw_model_steps or []
        self.config_steps = config_steps or []
        self.final_steps = final_steps or []

    def __add__(self, other):
        raw_model_steps = self.raw_model_steps + other.raw_model_steps
        config_steps = self.config_steps + other.config_steps
        final_steps = self.final_steps + other.final_steps

        result = FitPredictorSteps(raw_model_steps=raw_model_steps, config_steps=config_steps, final_steps=final_steps)

        return result


# noinspection PyMethodMayBeStatic
class FitPredictor(ABC):
    """
    Responsible for fit&predict&evaluation processes.

    Init parameters
    ----------
    raw_model_update_steps: List[RawModelUpdateStep]:
        Provide subclasses of RawModelUpdateStep to update your raw model
    config_update_steps: List[FitPredictConfigUpdateStep]
        Provide subclasses of FitPredictConfigUpdateStep to update different parameters
        in fit predict config
    final_steps: List[FinalHandlerStep]:
        Provide subclasses of FinalHandlerStep to perform any action using a resulting TravaModel
    logger: TravaLogger
        Logger object to inform you about the flow.
    """

    def __init__(self, steps: FitPredictorSteps = None, logger: TravaLogger = None):
        self._steps = steps or FitPredictorSteps()

        self._logger = logger or TravaLogger()

    def fit_predict(self, config: FitPredictConfig, tracker: Tracker) -> List[Evaluator]:
        """
        Prepares the data, performs one or many fit&predict, evaluates the results, tracks everything along the way.

        Parameters
        ----------
        config: FitPredictConfig
            Initial config object
        tracker: TravaTracker
            Helps with tracking things

        Returns
        -------
        List of Evaluator objects, one evaluator per fit.
        """
        raw_model = config.raw_model

        for updater in self._steps.raw_model_steps:
            raw_model = updater.update_model(raw_model=raw_model, config=config)

        # here will be stored evaluators for models, one for each model
        evaluators = []

        # each fit predictor can decide how many fits should we perform.
        # so subclasses prepare TravaModel and FitPredictorConfig for each fit.
        # yet using the same raw_model
        models_configs = self._models_configs(raw_model=raw_model, config=config)
        multiple_models_tracking = len(models_configs) > 1

        if multiple_models_tracking:
            # if we are going to train the raw model multiple times, we should group those fits
            # So we first track the "root" model, it will contain averaged results between all fits.
            self._start_tracking(
                config=config, raw_model=raw_model, model_id=config.model_id, tracker=tracker, nested=False
            )

        for trava_model, model_config in models_configs:
            # now fitting the model as many times as subclass wants
            model_id = trava_model.model_id
            split_result_fit = model_config.raw_split_data
            assert split_result_fit

            # here we start tracking one of the real fits.
            self._start_tracking(
                config=config, raw_model=raw_model, model_id=model_id, tracker=tracker, nested=multiple_models_tracking
            )

            fit_params = model_config.fit_params
            predict_params = model_config.predict_params
            # preparing data and parameters for fit&predict
            for config_updater in self._steps.config_steps:
                split_result_fit = config_updater.fit_split_data(
                    raw_split_data=split_result_fit, config=model_config, tracker=tracker
                )

                fit_params = config_updater.fit_params(
                    fit_params=fit_params, fit_split_data=split_result_fit, config=model_config, tracker=tracker
                )
                predict_params = config_updater.predict_params(
                    predict_params=predict_params, fit_split_data=split_result_fit, config=model_config, tracker=tracker
                )

            tracker.track_fit_params(model_id=trava_model.model_id, params=fit_params)
            tracker.track_predict_params(model_id=trava_model.model_id, params=predict_params)

            self._log_fit_start(trava_model=trava_model, split_result=split_result_fit)

            self._fit(
                trava_model=trava_model,
                X=split_result_fit.X_train,
                y=split_result_fit.y_train,
                fit_params=fit_params,
                predict_params=predict_params,
            )

            self._predict(trava_model=trava_model, X=split_result_fit.X_test, y=split_result_fit.y_test)

            # performing custom final steps provided in init
            for step in self._steps.final_steps:
                step.handle(trava_model=trava_model, config=model_config, tracker=tracker)

            self._logger.log(msg=f"Model evaluation {model_id}")

            # calculating metrics for the model
            evaluator = self._evaluator(model_config=model_config, split_result=split_result_fit, model=trava_model)
            evaluator.evaluate(scorers_providers=model_config.scorers_providers)

            self._track_metrics(model_id=model_id, evaluators=[evaluator], tracker=tracker)

            if model_config.serializer:
                self._logger.log(msg=f"Model serialization {model_id}")
                tracker.track_model_artifact(
                    model_id=model_id, model=trava_model.raw_model, serializer=model_config.serializer
                )

            evaluators.append(evaluator)
            tracker.end_tracking(model_id=model_id)

        if multiple_models_tracking:
            # just as we started tracking the "root" model in the beginning, here we have to end tracking
            self._track_metrics(model_id=config.model_id, evaluators=evaluators, tracker=tracker)
            tracker.end_tracking(model_id=config.model_id)

        return evaluators

    def _log_fit_start(self, trava_model: TravaModel, split_result: SplitResult):
        msg = f"Model fit start {trava_model.model_id}:\n"
        msg += f"train set shape ({split_result.X_train.shape})\n"
        if trava_model.is_classification_model:
            msg += f"target distribution ({dict(Counter(split_result.y_train))})\n"
        msg += f"test set shape ({split_result.X_test.shape})\n"
        if trava_model.is_classification_model:
            msg += f"target distribution ({dict(Counter(split_result.y_test))})\n"

        if split_result.X_valid is not None:
            msg += f"validation set shape ({split_result.X_valid.shape})\n"
            if trava_model.is_classification_model:
                msg += f"target distribution ({dict(Counter(split_result.y_valid))})\n"
        self._logger.log(msg=msg)

    def _evaluator(self, model_config: FitPredictConfig, split_result: SplitResult, model: TravaModel):
        """
        The method was extracted for unit-test purposes. :]
        """
        evaluator = Evaluator(
            trava_model=model, fit_split_data=split_result, raw_split_data=model_config.raw_split_data
        )
        return evaluator

    def _start_tracking(self, config: FitPredictConfig, raw_model, model_id: str, tracker: Tracker, nested: bool):
        tracker.start_tracking(model_id=model_id, nested=nested)
        self._track_model(
            model=raw_model,
            model_init_params=config.model_init_params,
            model_id=model_id,
            tracker=tracker,
            description=config.description,
        )

    def _track_metrics(self, model_id: str, evaluators: List[Evaluator], tracker: Tracker):
        model_results = ModelResult(model_id=model_id, evaluators=evaluators)
        tracker.track_model_results(model_results=model_results)

    def _track_model(self, model, model_init_params: dict, model_id: str, description: Optional[str], tracker: Tracker):
        tracker.track_model_init_params(model_id=model_id, params=model_init_params)
        tracker.track_model_info(model_id=model_id, model=model)
        if description:
            tracker.track_model_description(model_id=model_id, description=description)

    # TO OVERRIDE IF NEEDED

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        """
        If you want to run multiple fits on the same raw_model,
        just configure TravaModels and configs for them in your subclass.
        """
        return [(TravaModel(raw_model=raw_model, model_id=config.model_id), config)]

    def _fit(self, trava_model: TravaModel, X, y, fit_params: dict, predict_params: dict):
        """
        If you want to control the fit process
        """
        trava_model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    def _predict(self, trava_model: TravaModel, X, y):
        """
        If you want to control the predict process
        """
        trava_model.predict(X=X, y=y)



from abc import ABC, abstractmethod
from typing import Optional, List

from trava.split.logic import SplitLogic
from trava.split.result import SplitResult


class SplitResultHandler(ABC):
    """
    Can be used to modify SplitResult ( e.g. apply resampling )
    """

    @abstractmethod
    def handle(self, split_result: SplitResult) -> SplitResult:
        pass


class DataSplitConfig:
    """
    Data class containing all the data needed to make a split.
    """

    def __init__(
        self,
        split_logic: SplitLogic,
        target_col_name: str,
        test_size: float,
        valid_size: float = 0.0,
        split_result_handlers: Optional[List[SplitResultHandler]] = None,
        ignore_cols: Optional[List[str]] = None,
    ):
        self.split_logic = split_logic
        self.target_col_name = target_col_name
        self.test_size = test_size
        self.valid_size = valid_size
        self.split_result_handlers = split_result_handlers or []
        self.ignore_cols = ignore_cols or []



import copy
from typing import Optional, List, Type

from trava.scorers_provider import ScorersProvider
from trava.evaluator import Evaluator
from trava.fit_predictor import FitPredictor, FitPredictConfig
from trava.model_results import ModelResult
from trava.model_serializer import ModelSerializer
from trava.results_handler import ResultsHandler
from trava.scorer import Scorer
from trava.split.result import SplitResult
from trava.trava_tracker import TravaTracker
from trava.trava_base import _TravaBase


# noinspection PyPep8Naming
class TravaSV(_TravaBase):
    """
    Made for working with supervised problems.
    """

    def fit_predict(
        self,
        model_id: str,
        model_type: Type,
        description: Optional[str] = None,
        model_init_params: Optional[dict] = None,
        raw_split_data: Optional[SplitResult] = None,
        fit_predictor: FitPredictor = None,
        fit_params: dict = None,
        predict_params: dict = None,
        only_calculate_metrics: bool = False,
        keep_models_in_memory: bool = True,
        keep_data_in_memory: bool = True,
        serializer: Optional[ModelSerializer] = None,
    ):
        """
        Calls model's fit and predict with the data provided, calculates metrics and stores them.
        model_type and model_init_params were separated for the sake of easy parameters tracking.

        Parameters
        ----------
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        model_type: type of sklearn-style model
            Type of model that supports fit, predict and predict_proba methods
        description: str
            Describe the fit. It will be tracked if you set up tracker.
        model_init_params: dict
            Parameters to use to initialize model_type
        raw_split_data: optional SplitResult
             Already split train/test sets.
             Optional makes sense only if your FitPredictor contains
             some custom logic like CVFitPredictor.
        fit_predictor: FitPredictor
            Object responsible for performing fit and predict on a model
        fit_params: dict
            Custom params to use when calling model's fit method
        predict_params: dict
            Custom params to use when calling model's predict method
        only_calculate_metrics: bool
            if True, ResultsHandlers won't be called, but all the metrics
            for them will be ready. Can be useful when you run many
            consecutive runs and you don't to be overwhelmed by the output.
        keep_models_in_memory: bool
            Whether it's needed to store models in memory after the fit
        keep_data_in_memory: bool
            Whether it's needed to store the provided data in memory after the fit
        serializer: ModelSerializer
            If provided - fitted model will be serialized and tracked

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        fit_params = fit_params or {}
        predict_params = predict_params or {}
        model_init_params = model_init_params or {}
        fit_predictor = fit_predictor or FitPredictor(logger=self._logger)

        raw_model, model_init_params = self._create_raw_model(
            model_type=model_type, model_init_params=model_init_params
        )

        evaluators = self._fit_predict(
            raw_model=raw_model,
            model_id=model_id,
            model_init_params=model_init_params,
            fit_predictor=fit_predictor,
            fit_params=fit_params,
            predict_params=predict_params,
            serializer=serializer,
            split_result=raw_split_data,
            description=description,
        )
        result = None

        if only_calculate_metrics:
            self._prepare_model_results(evaluators=evaluators, save=True, main_model_id=model_id)
        else:
            result = self._results_for_evaluators(
                evaluators=evaluators, save_model_results=True, main_model_id=model_id
            )

        if not keep_models_in_memory:
            [evaluator.trava_model.unload_model() for evaluator in evaluators]

        if not keep_data_in_memory:
            [evaluator.unload_data() for evaluator in evaluators]
        return result

    def evaluate(self, model_id: str, results_handlers: List[ResultsHandler], save_results: bool) -> list:
        """
        Make possible to evaluate an existing model using new results handlers.
        Made in case you forgot to pass all the handlers in init
        or if you've come up with the new ideas.

        Note: in some cases it's impossible, see requires_raw_model/requires_X_y parameters in Scorer class.

        Parameters
        ----------
        model_id: str
            Unique identifier of the requested model
        results_handlers: list
            Each results handler provides logic how to present the metrics to a user.
        save_results: bool
            Whether you want to add the new metrics to the rest of them or just want to take a look.

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        temp_evaluators = []

        for evaluator in self._evaluators(model_id=model_id):
            if save_results:
                evaluator.evaluate(scorers_providers=results_handlers)

            # copy of evaluator just to get the results
            temp_evaluator = Evaluator(
                trava_model=evaluator.trava_model,
                fit_split_data=evaluator.fit_split_data,
                raw_split_data=evaluator.raw_split_data,
            )
            scorers_providers: List[ScorersProvider] = list(results_handlers)
            temp_evaluator.evaluate(scorers_providers=scorers_providers)
            temp_evaluators.append(temp_evaluator)

        result = self._results_for_evaluators(
            evaluators=temp_evaluators, results_handlers=results_handlers, main_model_id=model_id
        )

        return result

    def evaluate_track(self, model_id: str, scorers: List[Scorer]):
        """
        Similar to evaluate method, but this time it allows you to track new metrics

        Note: in some cases it's impossible, see requires_raw_model/requires_X_y parameters in Scorer class.

        Parameters
        ----------
        model_id: str
            Unique identifier of the requested model
        scorers: List[Scorer]
            List of scorers you want to use to track new metrics.
        """
        # we don't want to modify scorers in existing tracker
        tracker = self._tracker_copy()
        tracker.add_scorers(scorers=scorers)

        evaluators = self._evaluators(model_id=model_id)
        for evaluator in evaluators:
            evaluator.evaluate(scorers_providers=[tracker])

            self._track_evaluators(evaluators=[evaluator], model_id=evaluator.model_id, tracker=tracker)

        if len(evaluators) > 1:
            self._track_evaluators(evaluators=evaluators, model_id=model_id, tracker=tracker)

    def _tracker_copy(self) -> TravaTracker:
        """
        Note: in some cases it's impossible, see requires_raw_model/requires_X_y parameters in Scorer class.
        """
        return copy.deepcopy(self._tracker)

    @staticmethod
    def _track_evaluators(evaluators: List[Evaluator], model_id: str, tracker: TravaTracker):
        tracker.start_tracking(model_id=model_id)
        main_model_results = ModelResult(model_id=model_id, evaluators=evaluators)
        tracker.track_model_results(model_results=main_model_results)
        tracker.end_tracking(model_id=model_id)

    def _evaluators(self, model_id: str):
        model_results = self._model_results(model_id=model_id)
        return model_results.evaluators

    def detailed_results_for(self, model_id: str) -> Optional[list]:
        """
        Presents results for specific model.
        It may be a single fit as well as multiple CV fits.
        In case of multiple fits every fit will be shown separately.

        Parameters
        ----------
        model_id: str
            Unique identifier of the requested model

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        maybe_model_results = self._results.get(model_id)

        if not maybe_model_results:
            raise AttributeError("Model with id {} is not found.".format(model_id))

        return self._results_for_evaluators(evaluators=maybe_model_results.evaluators)

    def _results_for_evaluators(
        self,
        evaluators: List[Evaluator],
        save_model_results: bool = False,
        results_handlers: Optional[List[ResultsHandler]] = None,
        main_model_id: Optional[str] = None,
    ) -> list:
        # if main_model_id is provided it means that all evaluators
        # are related to the one run and must be averaged
        evaluators_results = self._prepare_model_results(
            evaluators=evaluators, save=save_model_results, main_model_id=main_model_id
        )

        all_results_handlers = results_handlers or self._results_handlers
        return self._results_for(results=evaluators_results, results_handlers=all_results_handlers)

    def _prepare_model_results(
        self, evaluators: List[Evaluator], save: bool = False, main_model_id: Optional[str] = None
    ) -> List[ModelResult]:
        is_agg_result = len(evaluators) > 1 and main_model_id
        evaluators_results = []
        for evaluator in evaluators:
            model_results = ModelResult(model_id=evaluator.model_id, evaluators=[evaluator])
            evaluators_results.append(model_results)

            if not is_agg_result and save:
                self._results[model_results.model_id] = model_results
        if is_agg_result:
            model_id = str(main_model_id)
            main_model_result = ModelResult(model_id=model_id, evaluators=evaluators)
            evaluators_results = [main_model_result]
            if save:
                self._results[model_id] = main_model_result
        return evaluators_results

    def _fit_predict(
        self,
        model_id: str,
        raw_model,
        model_init_params: Optional[dict],
        fit_predictor: FitPredictor,
        fit_params: dict,
        predict_params: dict,
        serializer: Optional[ModelSerializer],
        split_result: Optional[SplitResult] = None,
        description: Optional[str] = None,
    ):
        all_results_handlers = self._results_handlers + [self._tracker]
        scorers_providers: List[ScorersProvider] = list(all_results_handlers)

        config = FitPredictConfig(
            raw_split_data=split_result,
            raw_model=raw_model,
            model_init_params=model_init_params,
            model_id=model_id,
            scorers_providers=scorers_providers,
            serializer=serializer,
            fit_params=fit_params,
            predict_params=predict_params,
            description=description,
        )

        evaluators = fit_predictor.fit_predict(config=config, tracker=self._tracker)

        return evaluators



from typing import Optional

from trava.ext.boosting_eval.boosting_logic import CommonBoostingEvalLogic


class XGBoostEvalLogic(CommonBoostingEvalLogic):
    def _best_iteration(self, model) -> int:
        return model.best_iteration

    def _evals_results(self, model) -> dict:
        return model.evals_result()

    def _train_metrics_key(self, model) -> Optional[str]:
        if self._n_eval_sets(model=model) == 1:
            return None

        return self._results_key(idx=0)

    def _eval_metrics_key(self, model) -> str:
        n_eval_sets = self._n_eval_sets(model=model)
        result = self._results_key(idx=n_eval_sets - 1)
        return result

    @staticmethod
    def _results_key(idx: int) -> str:
        return "validation_" + str(idx)


class XGBoostRankerEvalLogic(XGBoostEvalLogic):
    def _evals_results(self, model) -> dict:
        return model.evals_result

    @staticmethod
    def _results_key(idx: int) -> str:
        return "eval_" + str(idx)



from typing import Optional

from trava.ext.boosting_eval.boosting_logic import CommonBoostingEvalLogic


class LightGBMEvalLogic(CommonBoostingEvalLogic):
    def _best_iteration(self, model) -> int:
        return model.best_iteration_

    def _evals_results(self, model) -> dict:
        return model.evals_result_

    def _train_metrics_key(self, model) -> Optional[str]:
        if self._n_eval_sets(model=model) == 1:
            return None

        return "training"

    def _eval_metrics_key(self, model) -> str:
        return "valid_" + str(self._n_eval_sets(model=model) - 1)



import pytest

from trava.split.config import DataSplitConfig
from trava.split.splitter import Splitter


def _config(mocker, valid_size=0.0):
    split_logic = mocker.Mock()
    split_logic.split.return_value = [mocker.MagicMock(), mocker.MagicMock(), mocker.MagicMock()]
    config = DataSplitConfig(
        split_logic=split_logic,
        target_col_name="target",
        test_size=0.3,
        valid_size=valid_size,
        split_result_handlers=[],
        ignore_cols=["col1", "col2"],
    )
    return config


@pytest.fixture(scope="class")
def additional_params():
    additional_params = {"param1": "value", "param2": 1}
    return additional_params


def test_split_logic(mocker, additional_params):
    df = mocker.Mock()
    config = _config(mocker=mocker)
    Splitter.split(df=df, config=config, **additional_params)

    config.split_logic.split.assert_called_once_with(
        data=df, test_size=config.test_size, valid_size=config.valid_size, **additional_params
    )


@pytest.mark.parametrize("n_handlers", [3])
def test_result_handlers(mocker, n_handlers):
    result_handlers = []
    config = _config(mocker=mocker)
    split_result = mocker.Mock()

    for _ in range(n_handlers):
        handler = mocker.Mock()
        handler.handle.return_value = split_result
        result_handlers.append(handler)

    config.split_result_handlers = result_handlers

    Splitter.split(df=mocker.Mock(), config=config)

    [handler.handle.assert_called_once() for handler in result_handlers]


@pytest.mark.parametrize("valid_size", [0.0, 0.2])
def test_split_X_y(mocker, additional_params, valid_size):
    df = mocker.Mock()
    config = _config(mocker=mocker, valid_size=valid_size)
    Splitter.split(df=df, config=config, **additional_params)

    train_df, test_df, valid_df = config.split_logic.split()
    train_df.drop.assert_called_once_with(config.ignore_cols + [config.target_col_name], axis=1)

    if valid_size != 0.0:
        valid_df.drop.assert_called_once_with(config.ignore_cols + [config.target_col_name], axis=1)



import random
import matplotlib.pyplot as plt
from typing import List, Tuple
from dataclasses import dataclass

from trava.ext.results_handlers.scorer_plotter import ScorerPlotter
from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.scorer import Scorer
from trava.trava_tracker import Tracker


@dataclass
class PlotItem:
    """
    scorer: Scorer
        How to calculate the metric we want to visualize
    plotter: ScorerPlotter
        How to visualize it
    can_overlap: bool
        If it's okay to use one canvas for drawing metrics for many models.
        e.g. ROC AUC plot is a good example when we can compare many models on the same graph,
             But Confusion matrix can be used only separately
    """

    scorer: Scorer
    plotter: ScorerPlotter
    can_overlap: bool


class PlotHandler(ResultsHandler):
    """
    Plots metrics.
    """

    def __init__(self, plot_items: List[PlotItem]):
        """
        Parameters
        ----------
        plot_items: list of PlotItem
            Each item in the list describes what & how to plot
        """
        super().__init__(scorers=[item.scorer for item in plot_items])

        self._plot_items = plot_items

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._show(results=results, logger=logger, tracker=tracker)
        self._track(results=results, logger=logger, tracker=tracker)

    def _show(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._plot(results=results, logger=logger, show=True, use_one_figure=True, tracker=tracker)

    def _track(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        self._plot(results=results, logger=logger, show=False, use_one_figure=False, tracker=tracker)

    def _plot(
        self,
        results: List[ModelResult],
        logger: TravaLogger,
        show: bool,
        tracker: Tracker,
        use_one_figure: bool = False,
        model_id: str = None,
    ):
        all_train_metrics: List[List[Metric]] = []
        all_test_metrics: List[List[Metric]] = []
        for model_result in results:
            if not model_result.is_one_fit_result:
                if show:
                    continue

                many_model_results = []
                for evaluator in model_result.evaluators:
                    many_model_results.append(ModelResult(model_id=evaluator.model_id, evaluators=[evaluator]))

                self._plot(
                    results=many_model_results,
                    logger=logger,
                    show=False,
                    tracker=tracker,
                    use_one_figure=True,
                    model_id=model_result.model_id,
                )

                self._plot(results=many_model_results, logger=logger, show=False, tracker=tracker, use_one_figure=False)
                continue
            if model_result.train_metrics(provider=self):
                model_train_metrics = model_result.train_metrics(provider=self)
                all_train_metrics.append(model_train_metrics)

            model_test_metrics = model_result.test_metrics(provider=self)
            all_test_metrics.append(model_test_metrics)

        if len(all_train_metrics) > 0:
            # TODO: couldn't figure out how to fix this, postponed it
            train_metrics_set: List[Tuple[Metric]] = list(zip(*all_train_metrics))  # type: ignore
            self._plot_metrics_set(
                metrics_set=train_metrics_set,
                label="Train",
                show=show,
                use_one_figure=use_one_figure,
                tracker=tracker,
                model_id=model_id,
            )

        # TODO: couldn't figure out how to fix this, postponed it
        test_metrics_set: List[Tuple[Metric]] = list(zip(*all_test_metrics))  # type: ignore
        self._plot_metrics_set(
            metrics_set=test_metrics_set,
            label="Test",
            show=show,
            use_one_figure=use_one_figure,
            tracker=tracker,
            model_id=model_id,
        )

    def _plot_metrics_set(
        self,
        metrics_set: List[Tuple[Metric]],
        label: str,
        show: bool,
        use_one_figure: bool,
        tracker: Tracker,
        model_id: str = None,
    ):
        if show:
            self._show_metrics_set(metrics_set=metrics_set, label=label, tracker=tracker)
        else:
            self._track_metrics_set(
                metrics_set=metrics_set, label=label, tracker=tracker, use_one_figure=use_one_figure, model_id=model_id
            )

    def _show_metrics_set(self, metrics_set: List[Tuple[Metric]], label: str, tracker: Tracker):
        fig, ax = self._fig_ax()
        for plot_idx, (metrics, plot_func) in self._enumerated_metrics_plots(metrics_set=metrics_set):
            self._plot_metrics(
                metrics=metrics,
                plot_item=plot_func,
                label=label,
                tracker=tracker,
                show=True,
                use_one_figure=True,
                fig=fig,
                ax=ax,
            )

    def _track_metrics_set(
        self, metrics_set: List[Tuple[Metric]], label: str, tracker: Tracker, use_one_figure: bool, model_id: str = None
    ):
        for plot_idx, (metrics, plot_item) in self._enumerated_metrics_plots(metrics_set=metrics_set):
            fig, ax = None, None
            if use_one_figure:
                fig, ax = self._fig_ax()

            self._plot_metrics(
                metrics=metrics,
                plot_item=plot_item,
                label=label,
                tracker=tracker,
                show=False,
                use_one_figure=use_one_figure,
                model_id=model_id,
                fig=fig,
                ax=ax,
            )

    def _plot_metrics(
        self,
        metrics: Tuple[Metric],
        plot_item: PlotItem,
        label: str,
        tracker: Tracker,
        show: bool,
        use_one_figure: bool,
        model_id: str = None,
        fig=None,
        ax=None,
    ):
        plt.clf()

        def color_for(idx):
            r = lambda: random.randint(0, 255)
            base_colors = ["b", "g", "r", "c", "m", "y", "k"]
            if idx > len(base_colors) - 1:
                return "#%02X%02X%02X" % (r(), r(), r())

            return base_colors[idx]

        use_one_figure = use_one_figure and plot_item.can_overlap

        if use_one_figure:
            fig, ax = self._fig_ax(existing_fig=fig)

        for metric_idx, metric in enumerate(metrics):
            if not use_one_figure:
                fig, ax = self._fig_ax(existing_fig=fig)

            plot_item.plotter.plot(metric=metric, fig=fig, ax=ax, color=color_for(idx=metric_idx), label=label)
            if show and not use_one_figure:
                plt.show()

            if not show:
                filename = (label + "_" + metric.name).lower()
                tracker.track_plot(model_id=model_id or metric.model_id, fig=fig, filename=filename)
                plt.close(fig)

        if show and use_one_figure:
            plt.show()

        if fig:
            plt.close(fig)

    def _enumerated_metrics_plots(self, metrics_set: List[Tuple[Metric]]):
        return enumerate(zip(metrics_set, self._plot_items))

    @staticmethod
    def _fig_ax(existing_fig=None) -> tuple:
        if existing_fig:
            plt.close(existing_fig)

        return plt.subplots(figsize=(10, 10))



from abc import abstractmethod
from typing import Optional

import matplotlib.pyplot as plt

from trava.tracker import Tracker


class CommonBoostingEvalLogic:
    """
    Some models support evaluation with early-stopping ( like gradient boosting ).
    This class was written with boosting in mind. Probably it can be generalized
    thus it won't be the base class anymore.
    """

    def __init__(self, needs_plot: bool, eval_metric: str = None, early_stopping_rounds: Optional[int] = 10):
        """
        Parameters
        ----------
        needs_plot: bool
            Whether to plot the results of evaluation at the end.
        eval_metric: str
            What metric to use for evaluation
        early_stopping_rounds: int
            Number of rounds to wait before perform early stopping
        """
        self._needs_plot = needs_plot
        self._eval_metric = eval_metric
        self._early_stopping_rounds = early_stopping_rounds

    def setup_eval(self, fit_params, X_train, y_train, X_eval, y_eval) -> dict:
        """
        You must make all the preparations to setup evaluation.

        Parameters
        ----------
        fit_params: dict
            Custom params to use when calling model's fit method
        X_train: pandas dataframe
            Train features for the model
        y_train: numpy array, pandas series
            Train target for the model
        X_eval: pandas dataframe
            Eval features for the model
        y_eval: numpy array, pandas series
            Eval target for the model

        Returns
        -------
        Prepared fit_params
        """
        result = fit_params.copy()

        if self._eval_metric:
            result["eval_metric"] = self._eval_metric

        result["eval_set"] = self._eval_set(X_train=X_train, y_train=y_train, X_eval=X_eval, y_eval=y_eval)

        if self._early_stopping_rounds:
            result["early_stopping_rounds"] = self._early_stopping_rounds

        return result

    def _eval_set(self, X_train, y_train, X_eval, y_eval):
        eval_set = [(X_eval, y_eval)]

        if self._user_train_in_eval:
            eval_set = [(X_train, y_train)] + eval_set

        return eval_set

    def train_set_results(self, model) -> Optional[list]:
        """
        Gets resulting metrics of evaluation for train set.

        Parameters
        ----------
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.

        Returns
        -------
        List containing evaluation metrics for train set.
        """
        eval_results = self._evals_results(model=model)
        key = self._train_metrics_key(model=model)

        if not key:
            return None

        return eval_results[key][self._get_metric_name(model=model)]

    def eval_set_results(self, model) -> list:
        """
        Gets resulting metrics of evaluation for eval set.

        Parameters
        ----------
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.

        Returns
        -------
        List containing evaluation metrics for eval set.
        """
        eval_results = self._evals_results(model=model)
        key = self._eval_metrics_key(model=model)

        return eval_results[key][self._get_metric_name(model=model)]

    def plot_if_needed(self, model_id: str, model, tracker: Tracker):
        """
        If plot was requested during initialization, learning curves
        will be plotted when evaluation has finished.

        Parameters
        ----------
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.
        tracker: Tracker
            Tracker for saving plots & eval results
        """
        if self._needs_plot:
            metric_to_plot = self._get_metric_name(model=model)

            train_results = self.train_set_results(model=model)
            eval_results = self.eval_set_results(model=model)

            epochs = len(eval_results)
            x_axis = range(0, epochs)

            fig, ax = plt.subplots()
            if train_results is not None:
                ax.plot(x_axis, train_results, label="Train")

            ax.plot(x_axis, eval_results, label="Test")
            ax.legend()
            plt.ylabel(metric_to_plot)
            plt.title("{} {}".format("{} validation results".format(model_id), metric_to_plot))
            plt.show()

            plot_filename = "eval_plot" + "_" + model_id
            tracker.track_plot(model_id=model_id, fig=fig, filename=plot_filename)

    def track_eval_metrics(self, model_id: str, model, tracker: Tracker):
        """
        Tracks eval results

        Parameters
        ----------
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.
        tracker: Tracker
            Tracker for saving plots & eval results
        """
        metric_to_plot = self._get_metric_name(model=model)

        train_results = self.train_set_results(model=model) or []
        eval_results = self.eval_set_results(model=model) or []

        tracker.track_tag(model_id=model_id, tag_key="eval_metric", tag_value=metric_to_plot)
        tracker.track_metric_value(model_id=model_id, name="early_stopping_rounds", value=self._early_stopping_rounds)
        tracker.track_metric_value(model_id=model_id, name="stopped_iteration", value=len(eval_results) - 1)
        tracker.track_metric_value(model_id=model_id, name="best_iteration", value=self._best_iteration(model=model))

        self._track_eval_results(
            model_id=model_id, eval_results=train_results, metric=metric_to_plot, train=True, tracker=tracker
        )

        self._track_eval_results(
            model_id=model_id, eval_results=eval_results, metric=metric_to_plot, train=False, tracker=tracker
        )

    @staticmethod
    def _track_eval_results(model_id: str, eval_results: list, metric: str, train: bool, tracker: Tracker):
        metric_name = ("train" if train else "eval") + "_" + metric

        for idx, value in enumerate(eval_results):
            tracker.track_metric_value(model_id=model_id, name=metric_name, value=value, step=idx)

    @abstractmethod
    def _best_iteration(self, model) -> int:
        pass

    @abstractmethod
    def _evals_results(self, model) -> dict:
        pass

    @abstractmethod
    def _train_metrics_key(self, model) -> Optional[str]:
        pass

    @abstractmethod
    def _eval_metrics_key(self, model) -> str:
        pass

    @property
    def _user_train_in_eval(self) -> bool:
        return True

    def _get_metric_name(self, model) -> str:
        eval_results = self._evals_results(model=model)
        result = list(eval_results[self._eval_metrics_key(model=model)].keys())[0]
        return result

    def _n_eval_sets(self, model) -> int:
        eval_results = self._evals_results(model=model)

        if not isinstance(eval_results, dict):
            raise ValueError("Eval results must be dict")

        n_eval_sets = len(eval_results.keys())
        assert 0 < n_eval_sets <= 2, "only train and test are now supported"

        return n_eval_sets



import pytest

from trava.fit_predictor import FitPredictConfig


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


@pytest.fixture()
def model(mocker, model_id):
    result = mocker.Mock()
    result.model_id = model_id
    return result


def test_equality(mocker, model, model_id):
    model_init_params = {"a": 1, "b": 2}
    scorers_providers = mocker.MagicMock()
    serializer = mocker.Mock()
    split_result = mocker.Mock()
    fit_params = {"c": 3, "d": 4}
    predict_params = {"_": 11, "+": 43}
    description = "descr"

    first_config = FitPredictConfig(
        raw_model=model,
        model_init_params=model_init_params,
        model_id=model_id,
        scorers_providers=scorers_providers,
        serializer=serializer,
        raw_split_data=split_result,
        fit_params=fit_params,
        predict_params=predict_params,
        description=description,
    )

    second_config = FitPredictConfig(
        raw_model=model,
        model_init_params=model_init_params,
        model_id=model_id,
        scorers_providers=scorers_providers,
        serializer=serializer,
        raw_split_data=split_result,
        fit_params=fit_params,
        predict_params=predict_params,
        description=description,
    )

    assert first_config == second_config
    second_config.serializer = mocker.Mock()
    assert first_config != second_config
    assert first_config != "str"

    none_config = FitPredictConfig(
        raw_model=None,
        model_init_params=None,
        model_id="123",
        scorers_providers=[],
        serializer=None,
        raw_split_data=mocker.Mock(),
    )
    assert first_config != none_config



from logging import Logger, INFO
from typing import Optional


class TravaLogger:
    """
    Wrapper of built-in python logger to have a control over it in future.
    """

    def __init__(self, logger: Optional[Logger] = None):
        self._logger = logger

    def log(self, msg: str, level=INFO):
        if self._logger:
            self._logger.log(msg=msg, level=level)



from typing import List, Callable

from trava.logger import TravaLogger
from trava.model_serializer import ModelSerializer
from trava.scorer import Scorer, OtherScorer
from trava.model_info import ModelInfo
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import Tracker


class TestScorer(Scorer):
    def _make_scorer(self, score_func: Callable, **metrics_kwargs) -> Callable:
        def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            metric_value = score_func(X["f1"].values)
            return metric_value

        return scorer


class TestAnyScorer(OtherScorer):
    def __init__(self, model_func: Callable):
        def score_func(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            return model_func(model_info)

        super().__init__(score_func=score_func)


class TestResultsHandler(ResultsHandler):
    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        all_train_metrics = []
        all_test_metrics = []
        all_any_metrics = []
        for model_result in results:
            all_train_metrics += model_result.train_metrics(provider=self)
            all_test_metrics += model_result.test_metrics(provider=self)
            all_any_metrics += model_result.other_metrics(provider=self)

        result = [
            [metric.value for metric in all_train_metrics],
            [metric.value for metric in all_test_metrics],
            [metric.value for metric in all_any_metrics] if all_any_metrics else [],
        ]
        return result


class TestModel:
    def __init__(self, required_param_1, required_param_2, def_1=1, def_2="tada", def_3=None, def_4=[], def_5=False):
        self._required_param_1 = required_param_1
        self._required_param_2 = required_param_2
        self._def_1 = def_1
        self._def_2 = def_2
        self._def_3 = def_3
        self._def_4 = def_4
        self._def_5 = def_5


class TestSerializer(ModelSerializer):
    def __init__(self):
        self._loaded = False
        self._saved = False

    def load(self, path: str):
        self._loaded = True

    def save(self, model, path: str):
        self._loaded = False



import xgboost as xgb
import os

from trava.model_serializer import ModelSerializer


class XGBoostModelSerializer(ModelSerializer):
    def load(self, path: str):
        model = xgb.Booster()
        model.load_model(os.path.abspath(path))

    def save(self, model, path: str):
        model.save_model(path)



from abc import ABC
from typing import Optional, List, Any, Dict

from trava.logger import TravaLogger
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import TravaTracker
from trava.utils.model_params_filter import merge_given_params_with_default


class _TravaBase(ABC):
    """
    Trava stands for TrainValidation.

    It will help you to gather metrics from different models you try
    and present them in any way that is convenient for you.

    Init parameters
    ----------
    logger: TravaLogger
        Encapsulates logic of work with native python logger
    tracker: Tracker
        Is used to track your experiments
    results_handlers: list
        Each results handler provides logic how to present the metrics to a user.
    """

    def __init__(
        self,
        logger: Optional[TravaLogger] = None,
        tracker: Optional[TravaTracker] = None,
        results_handlers: List[ResultsHandler] = None,
    ):
        self._logger = logger or TravaLogger()
        self._tracker: TravaTracker = tracker or TravaTracker(scorers=[])
        self._results_handlers = results_handlers or []

        self._results: Dict[str, ModelResult] = {}

    @property
    def results(self) -> list:
        """
        Calls every results handler to present all the metrics
        for all the fits that were made so far.

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        all_model_results = list(self._results.values())
        return self._results_for(results=all_model_results, results_handlers=self._results_handlers)

    def _results_for(self, results: List[ModelResult], results_handlers: List[ResultsHandler]) -> list:
        """
        Applies all the handlers and returns their result

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        result = []
        for handler in results_handlers:
            handler_result = handler.handle(results=results, logger=self._logger, tracker=self._tracker)
            if handler_result is not None:
                result.append(handler_result)

        return result

    def raw_models_for(self, model_id: str) -> Optional[Dict[str, Any]]:
        """
        Get raw models associated with the given model_id

        Parameters
        ----------
        model_id: str
            Unique identifier of the requested model

        Returns
        -------
        List of trained models if exists
        """
        model_results = self._model_results(model_id=model_id)
        return model_results.raw_models

    def _model_results(self, model_id: str) -> ModelResult:
        result = self._results.get(model_id)
        assert result, "You haven't fit a model with such model_id: {}".format(model_id)
        return result

    def _assert_existing_model(self, model_id: str):
        assert self._results.get(model_id) is None, "you tried to save a model with the already existing id"

    def _create_raw_model(self, model_type, model_init_params: dict) -> tuple:
        """
        Creates model object with the provided params.

        model_type: type of sklearn-style model
            Type of model that supports fit, predict and predict_proba methods
        model_init_params: dict
            Parameters to use to initialize model_type
        :return:
        Initialized model + trackable init parameters, including default ones.
        See _trackable_init_params_types for supported param types.
        """
        raw_model = model_type(**model_init_params)
        all_init_params = merge_given_params_with_default(object_type=model_type, params=model_init_params)

        return raw_model, all_init_params



import numpy as np
from sklearn.metrics import make_scorer
from typing import Callable

from trava.model_info import ModelInfo
from trava.scorer import Scorer


class SklearnScorer(Scorer):
    """
    Uses standard sklearn's make_scorer.
    """

    def __init__(
        self,
        score_func: Callable,
        greater_is_better=True,
        needs_proba=False,
        needs_threshold=False,
        sample_weight_required=False,
        threshold=None,
        **metrics_kwargs
    ):
        self._greater_is_better = greater_is_better
        self._needs_threshold = needs_threshold
        self._sample_weight_required = sample_weight_required
        self._threshold = threshold

        super().__init__(
            score_func=score_func,
            needs_proba=needs_proba,
            requires_raw_model=needs_threshold,
            requires_X_y=False,
            **metrics_kwargs
        )

    def _make_scorer(self, score_func: Callable, **metrics_kwargs) -> Callable:
        def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            if self._sample_weight_required and X is None:
                raise Exception(
                    "Sample weight is required for score ({}) calculation, "
                    "so it must be called with valid X_y data".format(self.func_name)
                )

            if X is None:
                # the last resort. If data and model are unloaded and your metric support interface
                # like (y_true, y_pred, **kwargs), then you will be able to calculate metrics

                y_cached = model_info.y(for_train=for_train)

                if self._needs_proba or self._threshold:
                    y_pred_values = model_info.y_pred_proba(for_train=for_train)

                    if len(set(y_cached)) == 2:
                        y_pred_values = y_pred_values[:, 1]

                        if self._threshold:
                            y_pred_values = np.where(y_pred_values > self._threshold, 1, 0)
                    elif self._threshold:
                        raise Exception("Only binary classification is supported for custom thresholds")
                else:
                    y_pred_values = model_info.y_pred(for_train=for_train)

                result = score_func(y_cached, y_pred_values, **metrics_kwargs)
            else:
                sk_scorer = self._get_sklearn_scorer(score_func=score_func, **metrics_kwargs)
                sample_weight = self._sample_weight(X=X, X_raw=X_raw)
                result = sk_scorer(model, X, y, sample_weight=sample_weight)
            return result

        return scorer

    def _get_sklearn_scorer(self, score_func: Callable, **metrics_kwargs):
        result = make_scorer(
            score_func=score_func,
            greater_is_better=self._greater_is_better,
            needs_proba=self._needs_proba,
            needs_threshold=self._needs_threshold,
            **metrics_kwargs
        )
        return result

    @staticmethod
    def _sample_weight(X, X_raw):
        """
        Maybe it will help to handle cases when the parameter is needed in a scorer.
        Introduced this template method just to not forget about it.
        """
        pass


# wrappers for sklearn metrics


def sk(score_func: Callable, **kwargs) -> Scorer:
    return _sk(score_func=score_func, needs_proba=False, **kwargs)


def sk_proba(score_func: Callable, **kwargs) -> Scorer:
    return _sk(score_func=score_func, needs_proba=True, **kwargs)


def _sk(score_func: Callable, needs_proba=False, **kwargs) -> Scorer:
    return SklearnScorer(score_func=score_func, needs_proba=needs_proba, **kwargs)



from collections import Counter

from trava.ext.boosting_eval.boosting_logic import CommonBoostingEvalLogic
from trava.ext.boosting_eval.eval_steps import EvalFitSteps
from trava.fit_predictor import FitPredictConfig, FitPredictConfigUpdateStep, FitPredictorSteps
from trava.split.result import SplitResult
from trava.tracker import Tracker


class _GroupConfigUpdateStep(FitPredictConfigUpdateStep):
    def __init__(self, group_col_name: str):
        self._group_col_name = group_col_name

    def fit_split_data(self, raw_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker) -> SplitResult:
        X_valid = None
        if raw_split_data.X_valid is not None:
            X_valid = raw_split_data.X_valid.drop(self._group_col_name, axis=1)

        result = SplitResult(
            X_train=raw_split_data.X_train.drop(self._group_col_name, axis=1),
            y_train=raw_split_data.y_train,
            X_test=raw_split_data.X_test.drop(self._group_col_name, axis=1),
            y_test=raw_split_data.y_test,
            X_valid=X_valid,
            y_valid=raw_split_data.y_valid,
        )

        return result

    def fit_params(
        self, fit_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        raw_split_data = config.raw_split_data
        assert raw_split_data
        train_counted_groups = self._counted_groups(X=raw_split_data.X_train)
        fit_params["group"] = train_counted_groups

        return fit_params

    def _counted_groups(self, X):
        train_groups = X[self._group_col_name].values
        counted_groups = list(Counter(train_groups).values())
        return counted_groups


class _GroupEvalConfigUpdateStep(_GroupConfigUpdateStep):
    def __init__(self, group_col_name: str):
        super().__init__(group_col_name=group_col_name)

    def fit_params(
        self, fit_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        fit_params = super().fit_params(
            fit_params=fit_params, fit_split_data=fit_split_data, config=config, tracker=tracker
        )

        raw_split_data = config.raw_split_data
        assert raw_split_data
        assert raw_split_data.X_valid is not None, "X_valid set must be present to run evaluation"

        eval_counted_groups = self._counted_groups(X=raw_split_data.X_valid)
        fit_params["eval_group"] = [fit_params["group"], eval_counted_groups]

        return fit_params

    def _counted_groups(self, X):
        train_groups = X[self._group_col_name].values
        counted_groups = list(Counter(train_groups).values())
        return counted_groups


class GroupFitSteps(FitPredictorSteps):
    """
    Simple extension for problems that are based on groups ( e.g. ranking )
    that provides group parameter for training a model.

    Init parameters
    ----------
    group_col_name: str
        Which column is used to store groups
    """

    def __init__(self, group_col_name: str):
        group_config_step = _GroupConfigUpdateStep(group_col_name=group_col_name)
        super().__init__(config_steps=[group_config_step])


class GroupEvalFitSteps(EvalFitSteps):
    """
    Same as GroupFitSteps, but also adds some modifications to support evaluation.

    Init parameters
    ----------
    eval_logic: Eval
        Contains logic of how to perform evaluation on the model.
    group_col_name: str
        Which column is used to store groups
    """

    def __init__(self, eval_logic: CommonBoostingEvalLogic, group_col_name: str):
        group_eval_config_step = _GroupEvalConfigUpdateStep(group_col_name=group_col_name)

        super().__init__(eval_logic=eval_logic)
        self.config_steps.insert(0, group_eval_config_step)



import pytest
import numpy as np

from trava.trava_model import TravaModel


@pytest.fixture(scope="class")
def model_id():
    return "test_model_id"


@pytest.fixture
def raw_model(mocker):
    return mocker.Mock()


@pytest.fixture
def X(mocker):
    return mocker.Mock()


@pytest.fixture
def y(mocker):
    return mocker.Mock()


@pytest.fixture
def fit_params(mocker):
    return {"fit_param": 111}


@pytest.fixture
def predict_params(mocker):
    return {"predict_param": 222}


@pytest.mark.parametrize("use_existing_model", [True, False], ids=["existing model", "no existing model"])
@pytest.mark.parametrize("only_fit", [True, False], ids=["only fit", "no only fit"])
def test_copy(mocker, model_id, use_existing_model, only_fit):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    y_train = np.array([0, 0, 1])
    y_train_pred = np.array([1, 2, 3])
    y_train_pred_proba = np.array([3, 4, 5])
    y_test = np.array([-2, 3, 5])
    y_test_pred = np.array([6, 7, 8])
    y_test_pred_proba = np.array([9, 10, 11])

    fit_params = {"1": 2}
    predict_params = {"2": 3}
    fit_time = 123
    predict_time = 434

    model._y_train = y_train
    model._y_train_pred = y_train_pred
    model._y_train_pred_proba = y_train_pred_proba

    model._y_test = y_test
    model._y_test_pred = y_test_pred
    model._y_test_pred_proba = y_test_pred_proba

    model._fit_params = fit_params
    model._predict_params = predict_params
    model._fit_time = fit_time
    model._predict_time = predict_time

    model_copy_id = model_id + "_copy"
    existing_model = None
    existing_model_id = "existing_model"
    # what a mess... but should work
    if use_existing_model:
        existing_model = TravaModel(raw_model=raw_model, model_id=existing_model_id)
        model_copy = model.copy(existing_model=existing_model, only_fit=only_fit)
    else:
        model_copy = model.copy(model_id=model_copy_id, only_fit=only_fit)

    if use_existing_model:
        assert model_copy.model_id == existing_model_id
    else:
        assert model_copy.model_id == model_copy_id

    y = model.y(for_train=True)
    copy_y = model_copy.y(for_train=True)
    assert np.array_equal(y, copy_y)
    assert model.fit_params == model_copy.fit_params
    assert model.fit_time == model_copy.fit_time

    if use_existing_model:
        assert existing_model == model_copy
    else:
        assert existing_model != model_copy

    if only_fit:
        assert model_copy.y(for_train=False) is None
        assert model_copy.y_pred(for_train=False) is None
        assert model_copy.y_pred_proba(for_train=False) is None
        assert model_copy.predict_time is None
        assert model_copy.predict_params == {}
    else:
        assert np.array_equal(model.y(for_train=False), model_copy.y(for_train=False))
        assert np.array_equal(model.y_pred(for_train=False), model_copy.y_pred(for_train=False))
        assert np.array_equal(model.y_pred_proba(for_train=False), model_copy.y_pred_proba(for_train=False))
        assert model.predict_time == model_copy.predict_time
        assert model.predict_params == model_copy.predict_params


def test_model_id(mocker, model_id):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.model_id == model_id


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_get_model(mocker, raw_model, X, y, needs_proba):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.get_model(for_train=True) == raw_model
    assert model.get_model(for_train=False) == raw_model

    y_predict_proba = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = y_predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model.fit(X=X, y=y)
    model.predict(X=X, y=y)

    model.unload_model()

    train_cached_model = model.get_model(for_train=True)
    test_cached_model = model.get_model(for_train=False)

    assert train_cached_model != raw_model
    assert test_cached_model != raw_model

    assert train_cached_model.predict(X) == y_pred
    if needs_proba:
        assert train_cached_model.predict_proba(X) == y_predict_proba


@pytest.mark.parametrize("for_train", [True, False], ids=["for_train", "for_test"])
def test_get_model_unload(mocker, raw_model, for_train):
    trava_model = TravaModel(raw_model=raw_model, model_id=model_id)
    trava_model.unload_model()

    with pytest.raises(ValueError):
        trava_model.get_model(for_train=for_train)

    if for_train:
        y_pred_key = "_y_train_pred"
    else:
        y_pred_key = "_y_test_pred"

    y_pred_mock = mocker.Mock()
    mocker.patch.object(trava_model, y_pred_key, y_pred_mock)

    assert trava_model.get_model(for_train=for_train).predict(X=None) == y_pred_mock


def test_raw_model(mocker, raw_model, model_id):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.raw_model == raw_model


def test_fit_params(mocker, raw_model, model_id, X, y, fit_params, predict_params):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    assert model.fit_params == fit_params


def test_predict_params(mocker, raw_model, model_id, X, y, fit_params, predict_params):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    assert model.predict_params == predict_params


def test_fit_time(mocker, raw_model, model_id, X, y):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    assert not model.fit_time
    model.fit(X=X, y=y)
    assert model.fit_time


def test_predict_time(mocker, raw_model, model_id, X, y):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    assert not model.predict_time
    model.predict(X=X, y=y)
    assert model.predict_time


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_fit(mocker, raw_model, model_id, X, y, fit_params, predict_params, needs_proba):
    if needs_proba:
        predict_proba = mocker.Mock()
        raw_model.predict_proba.return_value = predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    raw_model.fit.assert_called_once_with(X, y, **fit_params)
    raw_model.predict.assert_called_once_with(X, **predict_params)

    if needs_proba:
        raw_model.predict_proba.assert_called_with(X, **predict_params)

    assert model.fit_time


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_predict(mocker, raw_model, model_id, X, y, fit_params, needs_proba):
    if needs_proba:
        predict_proba = mocker.Mock()
        raw_model.predict_proba.return_value = predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.predict(X=X, y=y)

    raw_model.predict.assert_called_once_with(X)

    if needs_proba:
        raw_model.predict_proba.assert_called_with(X)

    assert model.predict_time


@pytest.mark.parametrize("is_classification", [True, False], ids=["is_classification", "no_classification"])
def test_is_classification(mocker, raw_model, model_id, X, y, fit_params, is_classification):
    predict_proba = mocker.Mock()

    if is_classification:
        raw_model.predict_proba.return_value = predict_proba
    else:
        raw_model.predict_proba = None

    model = TravaModel(raw_model=raw_model, model_id=model_id)

    if is_classification:
        assert model.is_classification_model
    else:
        assert not model.is_classification_model


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_all_y(mocker, raw_model, model_id, X, y, fit_params, predict_params, needs_proba):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    predict_proba_train = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = predict_proba_train
    y_pred_train = mocker.Mock()
    raw_model.predict.return_value = y_pred_train
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    predict_proba_test = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = predict_proba_test
    y_pred_test = mocker.Mock()
    raw_model.predict.return_value = y_pred_test
    X_test = mocker.Mock()
    y_test = mocker.Mock()
    model.predict(X=X_test, y=y_test)

    assert model.y_pred(for_train=True) == y_pred_train
    assert model.y_pred(for_train=False) == y_pred_test
    assert model.y(for_train=True) == y
    assert model.y(for_train=False) == y_test
    if needs_proba:
        assert model.y_pred_proba(for_train=True) == predict_proba_train
        assert model.y_pred_proba(for_train=False) == predict_proba_test


def test_unload(mocker, model_id):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.unload_model()

    assert not model.raw_model



from trava.ext.boosting_eval.boosting_logic import CommonBoostingEvalLogic
from trava.fit_predictor import FitPredictConfig, FitPredictConfigUpdateStep, FinalHandlerStep, FitPredictorSteps
from trava.tracker import Tracker
from trava.trava_model import TravaModel
from trava.split.result import SplitResult


class EvalConfigStep(FitPredictConfigUpdateStep):
    """
    Is used for training model with evaluation sets.
    Was made for boosting models.

    Init parameters
    ----------
    X_eval: pandas dataframe
            Eval features for the model
    y_eval: numpy array, pandas series
        Eval target for the model
    eval_logic: Eval
        Contains logic of how to perform evaluation on the model.
    """

    def __init__(self, eval_logic: CommonBoostingEvalLogic):
        self._eval_logic = eval_logic

    def fit_params(
        self, fit_params: dict, fit_split_data: SplitResult, config: FitPredictConfig, tracker: Tracker
    ) -> dict:
        split_result = fit_split_data
        self._assert_eval(X_eval=split_result.X_valid, y_eval=split_result.y_valid)

        fit_params = self._eval_logic.setup_eval(
            fit_params=fit_params,
            X_train=split_result.X_train,
            y_train=split_result.y_train,
            X_eval=split_result.X_valid,
            y_eval=split_result.y_valid,
        )

        return fit_params

    @staticmethod
    def _assert_eval(X_eval, y_eval):
        assert_text = "Couldn't find eval data in split_result"
        assert X_eval is not None and y_eval is not None, assert_text


class EvalFinalStep(FinalHandlerStep):
    """
    Is used for training model with evaluation sets.
    Was made for boosting models.

    Init parameters
    ----------
    eval_logic: Eval
        Contains logic of how to perform evaluation on the model.
    """

    def __init__(self, eval_logic: CommonBoostingEvalLogic):
        self._eval_logic = eval_logic

    def handle(self, trava_model: TravaModel, config: FitPredictConfig, tracker: Tracker):
        self._eval_logic.plot_if_needed(model_id=trava_model.model_id, model=trava_model.raw_model, tracker=tracker)
        self._eval_logic.track_eval_metrics(model_id=trava_model.model_id, model=trava_model.raw_model, tracker=tracker)


class EvalFitSteps(FitPredictorSteps):
    """
    Is used for training model with evaluation sets.
    Was made for boosting models.

    Init parameters
    ----------
    eval_logic: Eval
        Contains logic of how to perform evaluation on the model.
    """

    def __init__(self, eval_logic: CommonBoostingEvalLogic):
        config_step = EvalConfigStep(eval_logic=eval_logic)
        final_step = EvalFinalStep(eval_logic=eval_logic)

        super().__init__(config_steps=[config_step], final_steps=[final_step])



import pytest

from tests.objects_for_tests import TestResultsHandler


def test_scorers_copy(mocker):
    scorers = [mocker.Mock(), mocker.Mock()]
    initial_scorers = scorers.copy()
    handler = TestResultsHandler(scorers=scorers)

    scorers.append(mocker.Mock())

    assert handler.scorers == initial_scorers



from typing import List, Tuple

from trava.raw_dataset import RawDataset

from trava.ext.cv.base import CV
from trava.fit_predictor import FitPredictor, FitPredictConfig, FitPredictorSteps
from trava.logger import TravaLogger
from trava.trava_model import TravaModel
from trava.split.result import SplitResult


class CVFitPredictor(FitPredictor):
    def __init__(
        self,
        cv: CV,
        raw_dataset: RawDataset,
        ignore_cols: List[str],
        groups=None,
        steps: FitPredictorSteps = None,
        logger: TravaLogger = None,
    ):
        super().__init__(steps=steps or FitPredictorSteps(), logger=logger)

        self._cv = cv
        self._raw_dataset = raw_dataset
        self._groups = groups
        self._ignore_cols = ignore_cols

    def _models_configs(self, raw_model, config: FitPredictConfig) -> List[Tuple[TravaModel, FitPredictConfig]]:
        result = []

        X = self._raw_dataset.X
        y = self._raw_dataset.y
        X_cleaned = X.drop(self._ignore_cols, axis=1)

        for fold_idx, (train_indices, test_indices) in enumerate(self._cv.split(X=X_cleaned, y=y, groups=self._groups)):
            X_train, y_train = X_cleaned.iloc[train_indices], y.iloc[train_indices]
            X_test, y_test = X_cleaned.iloc[test_indices], y.iloc[test_indices]

            split_result = SplitResult(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)

            fold_model_id = config.model_id + "_fold_{}".format(fold_idx + 1)
            model_config = FitPredictConfig(
                raw_split_data=split_result,
                raw_model=config.raw_model,
                model_init_params=config.model_init_params,
                model_id=fold_model_id,
                scorers_providers=config.scorers_providers,
                serializer=config.serializer,
                fit_params=config.fit_params,
                predict_params=config.predict_params,
            )

            trava_model = TravaModel(raw_model=raw_model, model_id=fold_model_id)

            result.append((trava_model, model_config))

        return result



from typing import List, Optional, Dict
import mlflow
import mlflow.sklearn

from trava.scorer import Scorer
from trava.trava_tracker import TravaTracker
from trava.utils.model_params_filter import filter_params


class _RunInfo:
    def __init__(self, run_id: str, run_name: str):
        self.run_id = run_id
        self.run_name = run_name


class MLFlowTracker(TravaTracker):
    def __init__(self, scorers: List[Scorer]):
        super().__init__(scorers=scorers)

        self._model_id_run_info_map: Dict[str, _RunInfo] = {}

    def __getattr__(self, called_method):
        result = getattr(mlflow, called_method)
        if callable(result):

            def wrapper(*args, **kwargs):
                return getattr(mlflow, called_method)(*args, **kwargs)

            return wrapper
        raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, called_method))

    def _start_tracking(self, model_id: str, track_name: Optional[str] = None, nested: bool = False):
        if self._try_start_existing_run(model_id=model_id, nested=nested):
            return

        run_name = track_name or model_id
        mlflow.start_run(run_name=run_name, nested=nested)  # type: ignore
        self._save_run_id(run_name=run_name, model_id=model_id)

    def _try_start_existing_run(self, model_id: str, nested: bool):
        maybe_run_info = self._model_id_run_info_map.get(model_id)

        if not maybe_run_info:
            return False

        mlflow.start_run(run_id=maybe_run_info.run_id, nested=nested)  # type: ignore

        return True

    def _end_tracking(self, model_id: str):
        mlflow.end_run()  # type: ignore

    def _track_set_tracking_group(self, group: str):
        mlflow.set_experiment(group)  # type: ignore

    def _track_model_description(self, model_id: str, description: str):
        self.track_tag(model_id=model_id, tag_key="description", tag_value=description)

    def _track_model_init_params(self, model_id: str, params: dict):
        mlflow.log_params(params)  # type: ignore

    def _track_fit_params(self, model_id: str, params: dict):
        prepared_fit_params = self._prepare_params(params=params, key_prefix="fit_param")
        mlflow.log_params(prepared_fit_params)  # type: ignore

    def _track_predict_params(self, model_id: str, params: dict):
        prepared_predict_params = self._prepare_params(params=params, key_prefix="predict_param")
        mlflow.log_params(prepared_predict_params)  # type: ignore

    def _track_metric_value(self, model_id: str, name: str, value, step=None):
        mlflow.log_metric(name, value, step=step)  # type: ignore

    def _track_model_info(self, model_id: str, model):
        self.track_tag(model_id=model_id, tag_key="model_type", tag_value=type(model).__name__)

    def _track_tag(self, model_id: str, tag_key: str, tag_value):
        mlflow.set_tag(tag_key, tag_value)  # type: ignore

    def _track_artifact(self, model_id: str, file_path):
        mlflow.log_artifact(local_path=file_path)  # type: ignore

    def _save_run_id(self, run_name: str, model_id: str):
        run_id = mlflow.active_run().info.run_id  # type: ignore
        self._model_id_run_info_map[model_id] = _RunInfo(run_id=run_id, run_name=run_name)

    def _prepare_params(self, params: dict, key_prefix: str) -> dict:
        filtered_params = filter_params(params=params)

        for key in list(filtered_params.keys()):
            filtered_params[f"{key_prefix}__{key}"] = filtered_params.pop(key)

        return filtered_params



import pytest  # noqa

from trava.fit_predictor import FitPredictorSteps


def test_addition(mocker):
    def gen_steps(n: int):
        return [mocker.Mock() for _ in range(n)]

    first_raw_model_steps = gen_steps(n=2)
    second_raw_model_steps = gen_steps(n=1)

    first_config_steps = gen_steps(n=5)
    second_config_steps = gen_steps(n=0)

    first_final_steps = gen_steps(n=1)
    second_final_steps = gen_steps(n=1)

    first = FitPredictorSteps(
        raw_model_steps=first_raw_model_steps, config_steps=first_config_steps, final_steps=first_final_steps
    )
    second = FitPredictorSteps(
        raw_model_steps=second_raw_model_steps, config_steps=second_config_steps, final_steps=second_final_steps
    )

    result = first + second

    assert result.raw_model_steps == (first_raw_model_steps + second_raw_model_steps)
    assert result.config_steps == (first_config_steps + second_config_steps)
    assert result.final_steps == (first_final_steps + second_final_steps)



from typing import List

from trava.logger import TravaLogger
from trava.metric import Metric
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import Tracker


class LoggerHandler(ResultsHandler):
    """
    Just logs all the requested metrics.
    """

    def handle(self, results: List[ModelResult], logger: TravaLogger, tracker: Tracker):
        logger.log("*** Logging: ***")
        for model_results in results:
            logger.log("* Results for {} model *".format(model_results.model_id))

            if model_results.train_metrics(provider=self):
                logger.log("Train metrics:")
                LoggerHandler._log_metrics(metrics=model_results.train_metrics(provider=self), logger=logger)

            logger.log("Test metrics:")
            LoggerHandler._log_metrics(metrics=model_results.test_metrics(provider=self), logger=logger)

            logger.log("Other metrics:")
            LoggerHandler._log_metrics(metrics=model_results.other_metrics(provider=self), logger=logger)
        logger.log("*** END ***\n")

    @staticmethod
    def _log_metrics(metrics: List[Metric], logger: TravaLogger):
        for metric in metrics:
            logger.log(str(metric.name) + ":")
            logger.log(str(metric.value))

        logger.log("\n")
